# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.layers import preprocessing


class RandomGaussianBlurTest(tf.test.TestCase):
    def test_return_shapes(self):
        layer = preprocessing.RandomGaussianBlur(kernel_size=(3, 7), factor=(0, 2))

        # RGB
        xs = tf.ones((2, 512, 512, 3))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 512, 512, 3])

        # greyscale
        xs = tf.ones((2, 512, 512, 1))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 512, 512, 1])

    def test_in_single_image(self):
        layer = preprocessing.RandomGaussianBlur(kernel_size=(3, 7), factor=(0, 2))

        # RGB
        xs = tf.cast(
            tf.ones((512, 512, 3)),
            dtype=tf.float32,
        )

        xs = layer(xs)
        self.assertEqual(xs.shape, [512, 512, 3])

        # greyscale
        xs = tf.cast(
            tf.ones((512, 512, 1)),
            dtype=tf.float32,
        )

        xs = layer(xs)
        self.assertEqual(xs.shape, [512, 512, 1])

    def test_non_square_images(self):
        layer = preprocessing.RandomGaussianBlur(kernel_size=(3, 7), factor=(0, 2))

        # RGB
        xs = tf.ones((2, 256, 512, 3))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 256, 512, 3])

        # greyscale
        xs = tf.ones((2, 256, 512, 1))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 256, 512, 1])

    def test_single_input_args(self):
        layer = preprocessing.RandomGaussianBlur(kernel_size=7, factor=2)

        # RGB
        xs = tf.ones((2, 512, 512, 3))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 512, 512, 3])

        # greyscale
        xs = tf.ones((2, 512, 512, 1))
        xs = layer(xs)
        self.assertEqual(xs.shape, [2, 512, 512, 1])

    def test_numerical(self):
        layer = preprocessing.RandomGaussianBlur(kernel_size=3, factor=(1.0, 1.0))

        xs = tf.expand_dims(
            tf.constant([[0, 0, 0], [0, 1, 0], [0, 0, 0]]),
            axis=-1,
        )

        xs = tf.expand_dims(xs, axis=0)

        # Result expected to be identical to gaussian blur kernel of
        # size 3x3 and factor=1.0
        result = tf.expand_dims(
            tf.constant(
                [
                    [0.07511361, 0.1238414, 0.07511361],
                    [0.1238414, 0.20417996, 0.1238414],
                    [0.07511361, 0.1238414, 0.07511361],
                ]
            ),
            axis=-1,
        )
        result = tf.expand_dims(result, axis=0)
        xs = layer(xs)

        self.assertAllClose(xs, result)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import numpy as np
import tensorflow as tf

from keras_cv import bounding_box
from keras_cv.metrics.coco import COCOMeanAveragePrecision

SAMPLE_FILE = os.path.dirname(os.path.abspath(__file__)) + "/sample_boxes.npz"

delta = 0.04


class MeanAveragePrecisionTest(tf.test.TestCase):
    """Numerical testing for COCOMeanAveragePrecision.

    Unit tests that test Keras COCO metric results against the known values of
    cocoeval.py.  The bounding boxes in sample_boxes.npz were given to
    cocoeval.py, which computed the following values:

    Average Precision  (AP) @[ IoU=0.50:0.95 | area=   all | maxDets=100 ] = 0.617
    Average Precision  (AP) @[ IoU=0.50      | area=   all | maxDets=100 ] = 1.000
    Average Precision  (AP) @[ IoU=0.75      | area=   all | maxDets=100 ] = 0.707
    Average Precision  (AP) @[ IoU=0.50:0.95 | area= small | maxDets=100 ] = 0.604
    Average Precision  (AP) @[ IoU=0.50:0.95 | area=medium | maxDets=100 ] = 0.626
    Average Precision  (AP) @[ IoU=0.50:0.95 | area= large | maxDets=100 ] = 0.610
    """

    def test_mean_average_precision_correctness_default(self):
        y_true, y_pred, categories = load_samples(SAMPLE_FILE)

        # Area range all
        mean_average_precision = COCOMeanAveragePrecision(
            bounding_box_format="xyxy",
            class_ids=categories + [1000],
            max_detections=100,
            num_buckets=1000,
        )

        mean_average_precision.update_state(y_true, y_pred)
        result = mean_average_precision.result().numpy()
        self.assertAlmostEqual(result, 0.617, delta=delta)

    def test_mean_average_precision_correctness_iou_05(self):
        y_true, y_pred, categories = load_samples(SAMPLE_FILE)

        mean_average_precision = COCOMeanAveragePrecision(
            bounding_box_format="xyxy",
            class_ids=categories + [1000],
            iou_thresholds=[0.5],
            max_detections=100,
        )

        mean_average_precision.update_state(y_true, y_pred)
        result = mean_average_precision.result().numpy()
        self.assertAlmostEqual(result, 1.0, delta=delta)

    def test_mean_average_precision_correctness_iou_75(self):
        y_true, y_pred, categories = load_samples(SAMPLE_FILE)

        mean_average_precision = COCOMeanAveragePrecision(
            bounding_box_format="xyxy",
            class_ids=categories + [1000],
            iou_thresholds=[0.75],
            max_detections=100,
        )

        mean_average_precision.update_state(y_true, y_pred)
        result = mean_average_precision.result().numpy()
        self.assertAlmostEqual(result, 0.707, delta=delta)

    def test_mean_average_precision_correctness_medium(self):
        y_true, y_pred, categories = load_samples(SAMPLE_FILE)

        mean_average_precision = COCOMeanAveragePrecision(
            bounding_box_format="xyxy",
            class_ids=categories + [1000],
            max_detections=100,
            area_range=(32**2, 96**2),
        )

        mean_average_precision.update_state(y_true, y_pred)
        result = mean_average_precision.result().numpy()
        self.assertAlmostEqual(result, 0.626, delta=delta)

    def test_mean_average_precision_correctness_large(self):
        y_true, y_pred, categories = load_samples(SAMPLE_FILE)

        mean_average_precision = COCOMeanAveragePrecision(
            bounding_box_format="xyxy",
            class_ids=categories + [1000],
            max_detections=100,
            area_range=(96**2, 1e5**2),
        )

        mean_average_precision.update_state(y_true, y_pred)
        result = mean_average_precision.result().numpy()
        self.assertAlmostEqual(result, 0.610, delta=delta)

    def test_mean_average_precision_correctness_small(self):
        y_true, y_pred, categories = load_samples(SAMPLE_FILE)

        mean_average_precision = COCOMeanAveragePrecision(
            bounding_box_format="xyxy",
            class_ids=categories + [1000],
            max_detections=100,
            area_range=(0, 32**2),
        )

        mean_average_precision.update_state(y_true, y_pred)
        result = mean_average_precision.result().numpy()
        self.assertAlmostEqual(result, 0.604, delta=delta)


def load_samples(fname):
    npzfile = np.load(fname)
    y_true = npzfile["arr_0"].astype(np.float32)
    y_pred = npzfile["arr_1"].astype(np.float32)

    y_true = bounding_box.convert_format(y_true, source="xyWH", target="xyxy")
    y_pred = bounding_box.convert_format(y_pred, source="xyWH", target="xyxy")

    categories = set(int(x) for x in y_true[:, :, 4].numpy().flatten())
    categories = [x for x in categories if x != -1]

    return y_true, y_pred, categories



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl.testing import parameterized

from keras_cv import core
from keras_cv.layers import preprocessing


class RandomHueTest(tf.test.TestCase, parameterized.TestCase):
    def test_preserves_output_shape(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomHue(factor=(0.3, 0.8), value_range=(0, 255))
        output = layer(image)

        self.assertEqual(image.shape, output.shape)
        self.assertNotAllClose(image, output)

    def test_adjust_no_op(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomHue(factor=(0.0, 0.0), value_range=(0, 255))
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjust_full_opposite_hue(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomHue(factor=(1.0, 1.0), value_range=(0, 255))
        output = layer(image)

        channel_max = tf.math.reduce_max(output, axis=-1)
        channel_min = tf.math.reduce_min(output, axis=-1)
        # Make sure the max and min channel are the same between input and output
        # In the meantime, and channel will swap between each other.
        self.assertAllClose(channel_max, tf.math.reduce_max(image, axis=-1))
        self.assertAllClose(channel_min, tf.math.reduce_min(image, axis=-1))

    @parameterized.named_parameters(
        ("025", 0.25), ("05", 0.5), ("075", 0.75), ("100", 1.0)
    )
    def test_adjusts_all_values_for_factor(self, factor):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = preprocessing.RandomHue(factor=(factor, factor), value_range=(0, 255))
        output = layer(image)
        self.assertNotAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = preprocessing.RandomHue(factor=(0.0, 0.0), value_range=(0, 255))
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

        layer = preprocessing.RandomHue(factor=(0.3, 0.8), value_range=(0, 255))
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_with_uint8(self):
        image_shape = (4, 8, 8, 3)
        image = tf.cast(tf.random.uniform(shape=image_shape) * 255.0, dtype=tf.uint8)

        layer = preprocessing.RandomHue(factor=(0.0, 0.0), value_range=(0, 255))
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

        layer = preprocessing.RandomHue(factor=(0.3, 0.8), value_range=(0, 255))
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_config(self):
        layer = preprocessing.RandomHue(factor=(0.3, 0.8), value_range=(0, 255))
        config = layer.get_config()
        self.assertTrue(isinstance(config["factor"], core.UniformFactorSampler))
        self.assertEqual(config["factor"].get_config()["lower"], 0.3)
        self.assertEqual(config["factor"].get_config()["upper"], 0.8)
        self.assertEqual(config["value_range"], (0, 255))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv import bounding_box


class PadBatchToShapeTestCase(tf.test.TestCase):
    def test_bounding_box_padding(self):
        bounding_boxes = [[1, 2, 3, 4], [5, 6, 7, 8]]
        target_shape = [3, 4]
        result = bounding_box.pad_batch_to_shape(bounding_boxes, target_shape)
        self.assertAllClose(result, [[1, 2, 3, 4], [5, 6, 7, 8], [-1, -1, -1, -1]])

        target_shape = [2, 5]
        result = bounding_box.pad_batch_to_shape(bounding_boxes, target_shape)
        self.assertAllClose(result, [[1, 2, 3, 4, -1], [5, 6, 7, 8, -1]])

        # Make sure to raise error if the rank is different between bounding_box and
        # target shape
        with self.assertRaisesRegex(ValueError, "Target shape should have same rank"):
            bounding_box.pad_batch_to_shape(bounding_boxes, [1, 2, 3])

        # Make sure raise error if the target shape is smaller
        target_shape = [3, 2]
        with self.assertRaisesRegex(
            ValueError, "Target shape should be larger than bounding box shape"
        ):
            bounding_box.pad_batch_to_shape(bounding_boxes, target_shape)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for KerasCV models."""

import tensorflow as tf
from absl.testing import parameterized
from tensorflow.keras import backend

from keras_cv.models import densenet

MODEL_LIST = [
    (densenet.DenseNet121, 1024),
    (densenet.DenseNet169, 1664),
    (densenet.DenseNet201, 1920),
]


class ApplicationsTest(tf.test.TestCase, parameterized.TestCase):
    def assertShapeEqual(self, shape1, shape2):
        self.assertEqual(tf.TensorShape(shape1), tf.TensorShape(shape2))

    @parameterized.parameters(*MODEL_LIST)
    def test_application_base(self, app, _):
        # Can be instantiated with default arguments
        model = app(
            input_shape=(224, 224, 3),
            include_top=True,
            num_classes=1000,
            include_rescaling=False,
            weights=None,
        )
        # Can be serialized and deserialized
        config = model.get_config()
        reconstructed_model = model.__class__.from_config(config)
        self.assertEqual(len(model.weights), len(reconstructed_model.weights))
        backend.clear_session()

    @parameterized.parameters(*MODEL_LIST)
    def test_application_with_rescaling(self, app, last_dim):
        output_shape = _get_output_shape(
            lambda: app(
                include_rescaling=True,
                include_top=False,
                weights=None,
            )
        )
        self.assertShapeEqual(output_shape, (None, None, None, last_dim))
        backend.clear_session()

    @parameterized.parameters(*MODEL_LIST)
    def test_application_notop(self, app, last_dim):
        output_shape = _get_output_shape(
            lambda: app(
                include_rescaling=False,
                include_top=False,
                weights=None,
            )
        )
        self.assertShapeEqual(output_shape, (None, None, None, last_dim))
        backend.clear_session()

    @parameterized.parameters(*MODEL_LIST)
    def test_application_pooling(self, app, last_dim):
        output_shape = _get_output_shape(
            lambda: app(
                input_shape=(224, 224, 3),
                include_rescaling=False,
                include_top=False,
                weights=None,
                pooling="avg",
            )
        )
        self.assertShapeEqual(output_shape, (None, last_dim))

    @parameterized.parameters(*MODEL_LIST)
    def test_application_variable_input_channels(self, app, last_dim):
        input_shape = (None, None, 1)
        output_shape = _get_output_shape(
            lambda: app(
                include_rescaling=False,
                weights=None,
                include_top=False,
                input_shape=input_shape,
            )
        )
        self.assertShapeEqual(output_shape, (None, None, None, last_dim))
        backend.clear_session()

        input_shape = (None, None, 4)
        output_shape = _get_output_shape(
            lambda: app(
                include_rescaling=False,
                weights=None,
                include_top=False,
                input_shape=input_shape,
            )
        )
        self.assertShapeEqual(output_shape, (None, None, None, last_dim))
        backend.clear_session()


def _get_output_shape(model_fn):
    model = model_fn()
    return model.output_shape


if __name__ == "__main__":
    tf.test.main()



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class StochasticDepth(tf.keras.layers.Layer):
    """
    Implements the Stochastic Depth layer. It randomly drops residual branches
    in residual architectures. It is used as a drop-in replacement for addition
    operation. Note that this layer DOES NOT drop a residual block across
    individual samples but across the entire batch.

    Reference:
        - [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
        - Docstring taken from [stochastic_depth.py](https://tinyurl.com/mr3y2af6)

    Args:
        rate: float, the probability of the residual branch being dropped.

    Usage:

    `StochasticDepth` can be used in a residual network as follows:
    ```python
    # (...)
    input = tf.ones((1, 3, 3, 1), dtype=tf.float32)
    residual = tf.keras.layers.Conv2D(1, 1)(input)
    output = keras_cv.layers.StochasticDepth()([input, residual])
    # (...)
    ```

    At train time, StochasticDepth returns:
    $$
    x[0] + b_l * x[1],
    $$
    where $b_l$ is a random Bernoulli variable with probability
    $P(b_l = 1) = rate$. At test time, StochasticDepth rescales the activations
    of the residual branch based on the drop rate ($rate$):
    $$
    x[0] + (1 - rate) * x[1]
    $$
    """

    def __init__(self, rate=0.5, **kwargs):
        super().__init__(**kwargs)
        self.rate = rate
        self.survival_probability = 1.0 - self.rate

    def call(self, x, training=None):
        if len(x) != 2:
            raise ValueError(
                f"""Input must be a list of length 2. """
                f"""Got input with length={len(x)}."""
            )

        shortcut, residual = x

        b_l = tf.keras.backend.random_bernoulli([], p=self.survival_probability)

        if training:
            return shortcut + b_l * residual
        else:
            return shortcut + self.survival_probability * residual

    def get_config(self):
        config = {"rate": self.rate}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

import tensorflow as tf

from keras_cv import bounding_box
from keras_cv.metrics.coco import utils
from keras_cv.utils import iou as iou_lib


class COCOMeanAveragePrecision(tf.keras.metrics.Metric):
    """COCOMeanAveragePrecision computes an approximation of MaP.

    Args:
        class_ids: The class IDs to evaluate the metric for.  To evaluate for
            all classes in over a set of sequentially labelled classes, pass
            `range(num_classes)`.
        bounding_box_format: Format of the incoming bounding boxes.  Supported values
            are "xywh", "center_xywh", "xyxy".
        iou_thresholds: IoU thresholds over which to evaluate the recall.  Must
            be a tuple of floats, defaults to [0.5:0.05:0.95].
        area_range: area range to constrict the considered bounding boxes in
            metric computation. Defaults to `None`, which makes the metric
            count all bounding boxes.  Must be a tuple of floats.  The first
            number in the tuple represents a lower bound for areas, while the
            second value represents an upper bound.  For example, when
            `(0, 32**2)` is passed to the metric, recall is only evaluated for
            objects with areas less than `32*32`.  If `(32**2, 1000000**2)` is
            passed the metric will only be evaluated for boxes with areas larger
            than `32**2`, and smaller than `1000000**2`.
        max_detections: number of maximum detections a model is allowed to make.
            Must be an integer, defaults to `100`.
        recall_thresholds: The list of thresholds to average over in the MaP
            computation.  List of floats.  Defaults to [0:.01:1].
        num_buckets: num_buckets is used to select the number of confidence
            buckets predictions are placed into.  Instead of computation MaP
            over each incrementally selected set of bounding boxes, we instead
            place them into buckets.  This makes distributed computation easier.
            Increasing buckets improves accuracy of the metric, while decreasing
            buckets improves performance.  This is a tradeoff you must weight
            for your use case.  Defaults to 10,000 which is sufficiently large
            for most use cases.

    Usage:

    COCOMeanAveragePrecision accepts two Tensors as input to it's
    `update_state()` method.  These Tensors represent bounding boxes in
    `corners` format.  Utilities to convert Tensors from `xywh` to `corners`
    format can be found in `keras_cv.utils.bounding_box`.

    Each image in a dataset may have a different number of bounding boxes,
    both in the ground truth dataset and the prediction set.  In order to
    account for this, you may either pass a `tf.RaggedTensor`, or pad Tensors
    with `-1`s to indicate unused boxes.  A utility function to perform this
    padding is available at
    `keras_cv.bounding_box.pad_batch_to_shape()`.

    ```python
    coco_map = keras_cv.metrics.COCOMeanAveragePrecision(
        bounding_box_format='xyxy',
        max_detections=100,
        class_ids=[1]
    )

    y_true = np.array([[[0, 0, 10, 10, 1], [20, 20, 10, 10, 1]]]).astype(np.float32)
    y_pred = np.array([[[0, 0, 10, 10, 1, 1.0], [5, 5, 10, 10, 1, 0.9]]]).astype(
        np.float32
    )
    coco_map.update_state(y_true, y_pred)
    coco_map.result()
    # 0.24752477
    ```
    """

    def __init__(
        self,
        class_ids,
        bounding_box_format,
        recall_thresholds=None,
        iou_thresholds=None,
        area_range=None,
        max_detections=100,
        num_buckets=10000,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # Initialize parameter values
        self.bounding_box_format = bounding_box_format
        self.iou_thresholds = iou_thresholds or [x / 100.0 for x in range(50, 100, 5)]
        self.area_range = area_range
        self.max_detections = max_detections
        self.class_ids = list(class_ids)
        self.recall_thresholds = recall_thresholds or [x / 100 for x in range(0, 101)]
        self.num_buckets = num_buckets

        self.num_iou_thresholds = len(self.iou_thresholds)
        self.num_class_ids = len(self.class_ids)

        if any([c < 0 for c in class_ids]):
            raise ValueError(
                "class_ids must be positive.  Got " f"class_ids={class_ids}"
            )

        self.ground_truths = self.add_weight(
            "ground_truths",
            shape=(self.num_class_ids,),
            dtype=tf.int32,
            initializer="zeros",
        )
        self.true_positive_buckets = self.add_weight(
            "true_positive_buckets",
            shape=(
                self.num_class_ids,
                self.num_iou_thresholds,
                self.num_buckets,
            ),
            dtype=tf.int32,
            initializer="zeros",
        )
        self.false_positive_buckets = self.add_weight(
            "false_positive_buckets",
            shape=(
                self.num_class_ids,
                self.num_iou_thresholds,
                self.num_buckets,
            ),
            dtype=tf.int32,
            initializer="zeros",
        )

    def reset_state(self):
        self.true_positive_buckets.assign(tf.zeros_like(self.true_positive_buckets))
        self.false_positive_buckets.assign(tf.zeros_like(self.false_positive_buckets))
        self.ground_truths.assign(tf.zeros_like(self.ground_truths))

    @tf.function()
    def update_state(self, y_true, y_pred, sample_weight=None):
        if sample_weight is not None:
            warnings.warn(
                "sample_weight is not yet supported in keras_cv COCO metrics."
            )
        y_true = tf.cast(y_true, self.compute_dtype)
        y_pred = tf.cast(y_pred, self.compute_dtype)

        if isinstance(y_true, tf.RaggedTensor):
            y_true = y_true.to_tensor(default_value=-1)
        if isinstance(y_pred, tf.RaggedTensor):
            y_pred = y_pred.to_tensor(default_value=-1)

        y_true = bounding_box.convert_format(
            y_true,
            source=self.bounding_box_format,
            target="xyxy",
            dtype=self.compute_dtype,
        )
        y_pred = bounding_box.convert_format(
            y_pred,
            source=self.bounding_box_format,
            target="xyxy",
            dtype=self.compute_dtype,
        )

        class_ids = tf.constant(self.class_ids, dtype=self.compute_dtype)
        iou_thresholds = tf.constant(self.iou_thresholds, dtype=self.compute_dtype)

        num_images = tf.shape(y_true)[0]

        y_pred = utils.sort_bounding_boxes(y_pred, axis=bounding_box.XYXY.CONFIDENCE)

        ground_truth_boxes_update = tf.zeros_like(self.ground_truths)
        true_positive_buckets_update = tf.zeros_like(self.true_positive_buckets)
        false_positive_buckets_update = tf.zeros_like(self.false_positive_buckets)

        for img in tf.range(num_images):
            ground_truths = utils.filter_out_sentinels(y_true[img])
            detections = utils.filter_out_sentinels(y_pred[img])

            if self.area_range is not None:
                ground_truths = utils.filter_boxes_by_area_range(
                    ground_truths, self.area_range[0], self.area_range[1]
                )
                detections = utils.filter_boxes_by_area_range(
                    detections, self.area_range[0], self.area_range[1]
                )

            if self.max_detections < tf.shape(detections)[0]:
                detections = detections[: self.max_detections]

            true_positives_update = tf.TensorArray(
                tf.int32, size=self.num_class_ids * self.num_iou_thresholds
            )
            false_positives_update = tf.TensorArray(
                tf.int32, size=self.num_class_ids * self.num_iou_thresholds
            )
            ground_truths_update = tf.TensorArray(tf.int32, size=self.num_class_ids)

            for c_i in tf.range(self.num_class_ids):
                category_id = class_ids[c_i]
                ground_truths_by_category = utils.filter_boxes(
                    ground_truths, value=category_id, axis=bounding_box.XYXY.CLASS
                )
                detections_by_category = utils.filter_boxes(
                    detections, value=category_id, axis=bounding_box.XYXY.CLASS
                )
                if self.max_detections < tf.shape(detections_by_category)[0]:
                    detections_by_category = detections_by_category[
                        : self.max_detections
                    ]

                ground_truths_update = ground_truths_update.write(
                    c_i, tf.shape(ground_truths_by_category)[0]
                )

                ious = iou_lib.compute_ious_for_image(
                    ground_truths_by_category, detections_by_category
                )

                for iou_i in tf.range(self.num_iou_thresholds):
                    iou_threshold = iou_thresholds[iou_i]
                    pred_matches = utils.match_boxes(ious, iou_threshold)

                    dt_scores = detections_by_category[:, bounding_box.XYXY.CONFIDENCE]

                    true_positives = pred_matches != -1
                    false_positives = pred_matches == -1

                    dt_scores_clipped = tf.clip_by_value(dt_scores, 0.0, 1.0)
                    # We must divide by 1.01 to prevent off by one errors.
                    confidence_buckets = tf.cast(
                        tf.math.floor(self.num_buckets * (dt_scores_clipped / 1.01)),
                        tf.int32,
                    )
                    true_positives_by_bucket = tf.gather_nd(
                        confidence_buckets, indices=tf.where(true_positives)
                    )
                    false_positives_by_bucket = tf.gather_nd(
                        confidence_buckets, indices=tf.where(false_positives)
                    )

                    true_positive_counts_per_bucket = tf.math.bincount(
                        true_positives_by_bucket,
                        minlength=self.num_buckets,
                        maxlength=self.num_buckets,
                    )
                    false_positives_counts_per_bucket = tf.math.bincount(
                        false_positives_by_bucket,
                        minlength=self.num_buckets,
                        maxlength=self.num_buckets,
                    )

                    true_positives_update = true_positives_update.write(
                        (self.num_iou_thresholds * c_i) + iou_i,
                        true_positive_counts_per_bucket,
                    )
                    false_positives_update = false_positives_update.write(
                        (self.num_iou_thresholds * c_i) + iou_i,
                        false_positives_counts_per_bucket,
                    )

            true_positives_update = tf.reshape(
                true_positives_update.stack(),
                (self.num_class_ids, self.num_iou_thresholds, self.num_buckets),
            )
            false_positives_update = tf.reshape(
                false_positives_update.stack(),
                (self.num_class_ids, self.num_iou_thresholds, self.num_buckets),
            )

            true_positive_buckets_update = (
                true_positive_buckets_update + true_positives_update
            )
            false_positive_buckets_update = (
                false_positive_buckets_update + false_positives_update
            )
            ground_truth_boxes_update = (
                ground_truth_boxes_update + ground_truths_update.stack()
            )

        self.ground_truths.assign_add(ground_truth_boxes_update)
        self.true_positive_buckets.assign_add(true_positive_buckets_update)
        self.false_positive_buckets.assign_add(false_positive_buckets_update)

    @tf.function()
    def result(self):
        true_positives = tf.cast(self.true_positive_buckets, self.dtype)
        false_positives = tf.cast(self.false_positive_buckets, self.dtype)
        ground_truths = tf.cast(self.ground_truths, self.dtype)

        true_positives_sum = tf.cumsum(true_positives, axis=-1)
        false_positives_sum = tf.cumsum(false_positives, axis=-1)

        present_categories = tf.math.reduce_sum(tf.cast(ground_truths != 0, tf.int32))

        if present_categories == 0:
            return 0.0

        recalls = tf.math.divide_no_nan(
            true_positives_sum, ground_truths[:, None, None]
        )
        precisions = true_positives_sum / (false_positives_sum + true_positives_sum)

        result = tf.TensorArray(
            tf.float32, size=self.num_class_ids * self.num_iou_thresholds
        )
        zero_pad = tf.zeros(shape=(1,), dtype=tf.float32)
        # so in this case this should be: [1, 1]
        for i in tf.range(self.num_class_ids):
            for j in tf.range(self.num_iou_thresholds):
                recalls_i = recalls[i, j]
                precisions_i = precisions[i, j]

                # recall threshold=0 finds the first bucket always
                # this is different from the original implementation because the
                # original implementation always has at least one bounding box
                # in the first bucket.
                #
                # as such, we need to mask out the buckets where there is at
                # least one bounding box  Therefore, we must filter out the
                # buckets where (precisions_i) is NaN, as that implies a divide
                # by zero.

                inds = tf.where(not tf.math.is_nan(precisions_i))
                recalls_i = tf.gather_nd(recalls_i, inds)
                precisions_i = tf.gather_nd(precisions_i, inds)

                inds = tf.searchsorted(
                    recalls_i, tf.constant(self.recall_thresholds), side="left"
                )

                # if searchsorted returns len(precisions)+1, we should return 0
                precisions_i = tf.concat([precisions_i, zero_pad], axis=-1)
                precision_per_recall_threshold = tf.gather(precisions_i, inds)

                result_ij = tf.math.reduce_mean(precision_per_recall_threshold, axis=-1)
                result = result.write(j + i * self.num_iou_thresholds, result_ij)

        result = tf.reshape(
            result.stack(), (self.num_class_ids, self.num_iou_thresholds)
        )
        result = tf.math.reduce_mean(result, axis=-1)
        result = tf.math.reduce_sum(result, axis=0) / tf.cast(
            present_categories, tf.float32
        )
        return result

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "class_ids": self.class_ids,
                "bounding_box_format": self.bounding_box_format,
                "recall_thresholds": self.recall_thresholds,
                "iou_thresholds": self.iou_thresholds,
                "area_range": self.area_range,
                "max_detections": self.max_detections,
                "num_buckets": self.num_buckets,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import preprocessing


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class RandomJpegQuality(BaseImageAugmentationLayer):
    """Applies Random Jpeg compression artifacts to an image.

    Performs the jpeg compression algorithm on the image.  This layer can used in order
    to ensure your model is robust to artifacts introduced by JPEG compresion.

    Args:
        factor: 2 element tuple or 2 element list.  During augmentation, a random number
        is drawn from the factor distribution.  This value is passed to
        `tf.image.adjust_jpeg_quality()`.
        seed: Integer. Used to create a random seed.

    Usage:
    ```python
    layer = keras_cv.RandomJpegQuality(factor=(75, 100)))
    (images, labels), _ = tf.keras.datasets.cifar10.load_data()
    augmented_images = layer(images)
    ```
    """

    def __init__(self, factor, seed=None, **kwargs):
        super().__init__(**kwargs)
        if isinstance(factor, (float, int)):
            raise ValueError(
                "RandomJpegQuality() expects factor to be a 2 element "
                "tuple, list or a `keras_cv.FactorSampler`. "
                "RandomJpegQuality() received `factor={factor}`."
            )
        self.seed = seed
        self.factor = preprocessing.parse_factor(
            factor, min_value=0, max_value=100, param_name="factor", seed=self.seed
        )

    def get_random_transformation(self, **kwargs):
        return self.factor(dtype=tf.int32)

    def augment_image(self, image, transformation=None, **kwargs):
        jpeg_quality = transformation
        return tf.image.adjust_jpeg_quality(image, jpeg_quality)

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    def get_config(self):
        config = super().get_config()
        config.update({"factor": self.factor})
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for util functions."""

import tensorflow as tf

from keras_cv import bounding_box
from keras_cv.metrics.coco import utils
from keras_cv.utils import iou as iou_lib


class UtilTest(tf.test.TestCase):
    def test_filter_bounding_boxes_empty(self):
        # set of bounding_boxes
        y_pred = tf.stack([_dummy_bounding_box(category=1)])
        result = utils.filter_boxes(y_pred, 2, axis=bounding_box.XYXY.CLASS)

        self.assertEqual(result.shape[0], 0)

    def test_bounding_box_area(self):
        boxes = tf.constant([[0, 0, 100, 100]], dtype=tf.float32)
        areas = utils.bounding_box_area(boxes)
        self.assertAllClose(areas, tf.constant((10000.0,)))

    def test_filter_bounding_boxes(self):
        # set of bounding_boxes
        y_pred = tf.stack(
            [_dummy_bounding_box(category=1), _dummy_bounding_box(category=2)]
        )
        result = utils.filter_boxes(y_pred, 2, axis=bounding_box.XYXY.CLASS)

        self.assertAllClose(result, tf.stack([_dummy_bounding_box(category=2)]))

    def test_to_sentinel_padded_bounding_box_tensor(self):
        box_set1 = tf.stack([_dummy_bounding_box(), _dummy_bounding_box()])
        box_set2 = tf.stack([_dummy_bounding_box()])
        boxes = [box_set1, box_set2]
        bounding_box_tensor = utils.to_sentinel_padded_bounding_box_tensor(boxes)
        self.assertAllClose(
            bounding_box_tensor[1, 1],
            -tf.ones(
                6,
            ),
        )

    def test_filter_out_sentinels(self):
        # set of bounding_boxes
        y_pred = tf.stack(
            [_dummy_bounding_box(category=1), _dummy_bounding_box(category=-1)]
        )
        result = utils.filter_out_sentinels(y_pred)

        self.assertAllClose(result, tf.stack([_dummy_bounding_box(category=1)]))

    def test_end_to_end_sentinel_filtering(self):
        box_set1 = tf.stack([_dummy_bounding_box(), _dummy_bounding_box()])
        box_set2 = tf.stack([_dummy_bounding_box()])
        boxes = [box_set1, box_set2]
        bounding_box_tensor = utils.to_sentinel_padded_bounding_box_tensor(boxes)

        self.assertAllClose(
            utils.filter_out_sentinels(bounding_box_tensor[0]), box_set1
        )
        self.assertAllClose(
            utils.filter_out_sentinels(bounding_box_tensor[1]), box_set2
        )

    def test_match_boxes(self):
        y_pred = tf.stack(
            [
                _dummy_bounding_box(0.1),
                _dummy_bounding_box(0.9),
                _dummy_bounding_box(0.4),
            ]
        )
        y_true = tf.stack(
            [
                _dummy_bounding_box(0.1),
                _dummy_bounding_box(0.9),
                _dummy_bounding_box(0.4),
                _dummy_bounding_box(0.2),
            ]
        )

        ious = iou_lib.compute_ious_for_image(y_true, y_pred)
        self.assertEqual(utils.match_boxes(ious, 0.5).shape, [3])

    def test_sort_bounding_boxes_unsorted_list(self):
        y_pred = tf.expand_dims(
            tf.stack(
                [
                    _dummy_bounding_box(0.1),
                    _dummy_bounding_box(0.9),
                    _dummy_bounding_box(0.4),
                    _dummy_bounding_box(0.2),
                ]
            ),
            axis=0,
        )
        want = tf.expand_dims(
            tf.stack(
                [
                    _dummy_bounding_box(0.9),
                    _dummy_bounding_box(0.4),
                    _dummy_bounding_box(0.2),
                    _dummy_bounding_box(0.1),
                ]
            ),
            axis=0,
        )
        y_sorted = utils.sort_bounding_boxes(y_pred, bounding_box.XYXY.CONFIDENCE)
        self.assertAllClose(y_sorted, want)

    def test_sort_bounding_boxes_empty_list(self):
        y_pred = tf.stack([])
        y_sorted = utils.sort_bounding_boxes(y_pred)
        self.assertAllClose(y_pred, y_sorted)


def _dummy_bounding_box(confidence=0.0, category=0):
    """returns a bounding_box dummy with all 0 values, except for confidence."""
    return tf.constant([0, 0, 0, 0, category, confidence])



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import preprocessing


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class RandomSaturation(BaseImageAugmentationLayer):
    """Randomly adjusts the saturation on given images.

    This layer will randomly increase/reduce the saturation for the input RGB
    images. At inference time, the output will be identical to the input.
    Call the layer with `training=True` to adjust the brightness of the input.

    Args:
        factor: A tuple of two floats, a single float or `keras_cv.FactorSampler`.
            `factor` controls the extent to which the image saturation is impacted.
            `factor=0.5` makes this layer perform a no-op operation. `factor=0.0` makes
            the image to be fully grayscale. `factor=1.0` makes the image to be fully
            saturated.
            Values should be between `0.0` and `1.0`. If a tuple is used, a `factor`
            is sampled between the two values for every image augmented.  If a single
            float is used, a value between `0.0` and the passed float is sampled.
            In order to ensure the value is always the same, please pass a tuple with
            two identical floats: `(0.5, 0.5)`.
        seed: Integer. Used to create a random seed.
    """

    def __init__(self, factor, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.factor = preprocessing.parse_factor(
            factor,
            min_value=0.0,
            max_value=1.0,
        )
        self.seed = seed

    def get_random_transformation(self, image=None, label=None, bounding_boxes=None):
        del image, label, bounding_boxes
        return self.factor()

    def augment_image(self, image, transformation=None, **kwargs):
        # Convert the factor range from [0, 1] to [0, +inf]. Note that the
        # tf.image.adjust_saturation is trying to apply the following math formula
        # `output_saturation = input_saturation * factor`. We use the following
        # method to the do the mapping.
        # `y = x / (1 - x)`.
        # This will ensure:
        #   y = +inf when x = 1 (full saturation)
        #   y = 1 when x = 0.5 (no augmentation)
        #   y = 0 when x = 0 (full gray scale)

        # Convert the transformation to tensor in case it is a float. When
        # transformation is 1.0, then it will result in to divide by zero error, but
        # it will be handled correctly when it is a one tensor.
        transformation = tf.convert_to_tensor(transformation)
        adjust_factor = transformation / (1 - transformation)
        return tf.image.adjust_saturation(image, saturation_factor=adjust_factor)

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    def get_config(self):
        config = {
            "factor": self.factor,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv import core
from keras_cv.layers import preprocessing


class RandomSaturationTest(tf.test.TestCase):
    def test_preserves_output_shape(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomSaturation(factor=(0.3, 0.8))
        output = layer(image)

        self.assertEqual(image.shape, output.shape)
        self.assertNotAllClose(image, output)

    def test_no_adjustment_for_factor_point_five(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomSaturation(factor=(0.5, 0.5))
        output = layer(image)

        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

    def test_adjust_to_grayscale(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomSaturation(factor=(0.0, 0.0))
        output = layer(image)

        channel_mean = tf.math.reduce_mean(output, axis=-1)
        channel_values = tf.unstack(output, axis=-1)
        # Make sure all the pixel has the same value among the channel dim, which is
        # a fully gray RGB.
        for channel_value in channel_values:
            self.assertAllClose(channel_mean, channel_value, atol=1e-5, rtol=1e-5)

    def test_adjust_to_full_saturation(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0

        layer = preprocessing.RandomSaturation(factor=(1.0, 1.0))
        output = layer(image)

        channel_mean = tf.math.reduce_min(output, axis=-1)
        # Make sure at least one of the channel is 0.0 (fully saturated image)
        self.assertAllClose(channel_mean, tf.zeros((4, 8, 8)))

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = preprocessing.RandomSaturation(factor=(0.5, 0.5))
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

        layer = preprocessing.RandomSaturation(factor=(0.3, 0.8))
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_with_unit8(self):
        image_shape = (4, 8, 8, 3)
        image = tf.cast(tf.random.uniform(shape=image_shape) * 255.0, dtype=tf.uint8)

        layer = preprocessing.RandomSaturation(factor=(0.5, 0.5))
        output = layer(image)
        self.assertAllClose(image, output, atol=1e-5, rtol=1e-5)

        layer = preprocessing.RandomSaturation(factor=(0.3, 0.8))
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_config(self):
        layer = preprocessing.RandomSaturation(factor=(0.3, 0.8))
        config = layer.get_config()
        self.assertTrue(isinstance(config["factor"], core.UniformFactorSampler))
        self.assertEqual(config["factor"].get_config()["lower"], 0.3)
        self.assertEqual(config["factor"].get_config()["upper"], 0.8)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DenseNet models for KerasCV.

Reference:
  - [Densely Connected Convolutional Networks](https://arxiv.org/abs/1608.06993)
  - [Based on the Original keras.applications DenseNet](https://github.com/keras-team/keras/blob/master/keras/applications/densenet.py)
"""

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend
from tensorflow.keras import layers

BN_AXIS = 3

BASE_DOCSTRING = """Instantiates the {name} architecture.

    Reference:
        - [Densely Connected Convolutional Networks (CVPR 2017)](https://arxiv.org/abs/1608.06993)

    This function returns a Keras {name} model.

    For transfer learning use cases, make sure to read the [guide to transfer
        learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Args:
        include_rescaling: whether or not to Rescale the inputs.If set to True,
            inputs will be passed through a `Rescaling(1/255.0)` layer.
        include_top: whether to include the fully-connected layer at the top of the
            network.  If provided, num_classes must be provided.
        num_classes: optional number of classes to classify images into, only to be
            specified if `include_top` is True, and if no `weights` argument is
            specified.
        weights: one of `None` (random initialization), or a pretrained weight file
            path.
        input_shape: optional shape tuple, defaults to (None, None, 3).
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be the 4D tensor output
                of the last convolutional block.
            - `avg` means that global average pooling will be applied to the output
                of the last convolutional block, and thus the output of the model will
                be a 2D tensor.
            - `max` means that global max pooling will be applied.
        name: (Optional) name to pass to the model.  Defaults to "{name}".

    Returns:
      A `keras.Model` instance.
"""


def DenseBlock(blocks, name=None):
    """A dense block.

    Args:
      blocks: integer, the number of building blocks.
      name: string, block label.

    Returns:
      a function that takes an input Tensor representing a DenseBlock.
    """
    if name is None:
        name = f"dense_block_{backend.get_uid('dense_block')}"

    def apply(x):
        for i in range(blocks):
            x = ConvBlock(32, name=f"{name}_block_{i}")(x)
        return x

    return apply


def TransitionBlock(reduction, name=None):
    """A transition block.

    Args:
      reduction: float, compression rate at transition layers.
      name: string, block label.

    Returns:
      a function that takes an input Tensor representing a TransitionBlock.
    """
    if name is None:
        name = f"transition_block_{backend.get_uid('transition_block')}"

    def apply(x):
        x = layers.BatchNormalization(
            axis=BN_AXIS, epsilon=1.001e-5, name=f"{name}_bn"
        )(x)
        x = layers.Activation("relu", name=f"{name}_relu")(x)
        x = layers.Conv2D(
            int(backend.int_shape(x)[BN_AXIS] * reduction),
            1,
            use_bias=False,
            name=f"{name}_conv",
        )(x)
        x = layers.AveragePooling2D(2, strides=2, name=f"{name}_pool")(x)
        return x

    return apply


def ConvBlock(growth_rate, name=None):
    """A building block for a dense block.

    Args:
      growth_rate: float, growth rate at dense layers.
      name: string, block label.

    Returns:
      a function that takes an input Tensor representing a ConvBlock.
    """
    if name is None:
        name = f"conv_block_{backend.get_uid('conv_block')}"

    def apply(x):
        x1 = layers.BatchNormalization(
            axis=BN_AXIS, epsilon=1.001e-5, name=f"{name}_0_bn"
        )(x)
        x1 = layers.Activation("relu", name=f"{name}_0_relu")(x1)
        x1 = layers.Conv2D(4 * growth_rate, 1, use_bias=False, name=f"{name}_1_conv")(
            x1
        )
        x1 = layers.BatchNormalization(
            axis=BN_AXIS, epsilon=1.001e-5, name=f"{name}_1_bn"
        )(x1)
        x1 = layers.Activation("relu", name=f"{name}_1_relu")(x1)
        x1 = layers.Conv2D(
            growth_rate, 3, padding="same", use_bias=False, name=f"{name}_2_conv"
        )(x1)
        x = layers.Concatenate(axis=BN_AXIS, name=f"{name}_concat")([x, x1])
        return x

    return apply


def DenseNet(
    blocks,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_shape=(None, None, 3),
    pooling=None,
    classifier_activation="softmax",
    name=None,
    **kwargs,
):
    """Instantiates the DenseNet architecture.

    Reference:
        - [Densely Connected Convolutional Networks (CVPR 2017)](https://arxiv.org/abs/1608.06993)

    This function returns a Keras DenseNet model.

    For transfer learning use cases, make sure to read the [guide to transfer
        learning & fine-tuning](https://keras.io/guides/transfer_learning/).

    Args:
        blocks: numbers of building blocks for the four dense layers.
        include_rescaling: whether or not to Rescale the inputs.If set to True,
            inputs will be passed through a `Rescaling(1/255.0)` layer.
        include_top: whether to include the fully-connected layer at the top of the
            network.  If provided, num_classes must be provided.
        num_classes: optional number of classes to classify images into, only to be
            specified if `include_top` is True, and if no `weights` argument is
            specified.
        weights: one of `None` (random initialization), or a pretrained weight file
            path.
        input_shape: optional shape tuple, defaults to (None, None, 3).
        pooling: optional pooling mode for feature extraction
            when `include_top` is `False`.
            - `None` means that the output of the model will be the 4D tensor output
                of the last convolutional block.
            - `avg` means that global average pooling will be applied to the output
                of the last convolutional block, and thus the output of the model will
                be a 2D tensor.
            - `max` means that global max pooling will be applied.
        classifier_activation: A `str` or callable. The activation function to use
            on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top" layer.
            When loading pretrained weights, `classifier_activation` can only
            be `None` or `"softmax"`.
        name: (Optional) name to pass to the model.  Defaults to "DenseNet".

    Returns:
      A `keras.Model` instance.
    """
    if weights and not tf.io.gfile.exists(weights):
        raise ValueError(
            "The `weights` argument should be either `None` or the path to the "
            "weights file to be loaded. Weights file not found at location: {weights}"
        )

    if include_top and not num_classes:
        raise ValueError(
            "If `include_top` is True, you should specify `num_classes`. "
            f"Received: num_classes={num_classes}"
        )

    inputs = layers.Input(shape=input_shape)

    x = inputs
    if include_rescaling:
        x = layers.Rescaling(1 / 255.0)(x)

    x = layers.Conv2D(
        64, 7, strides=2, use_bias=False, padding="same", name="conv1/conv"
    )(x)
    x = layers.BatchNormalization(axis=BN_AXIS, epsilon=1.001e-5, name="conv1/bn")(x)
    x = layers.Activation("relu", name="conv1/relu")(x)
    x = layers.MaxPooling2D(3, strides=2, padding="same", name="pool1")(x)

    x = DenseBlock(blocks[0], name="conv2")(x)
    x = TransitionBlock(0.5, name="pool2")(x)
    x = DenseBlock(blocks[1], name="conv3")(x)
    x = TransitionBlock(0.5, name="pool3")(x)
    x = DenseBlock(blocks[2], name="conv4")(x)
    x = TransitionBlock(0.5, name="pool4")(x)
    x = DenseBlock(blocks[3], name="conv5")(x)

    x = layers.BatchNormalization(axis=BN_AXIS, epsilon=1.001e-5, name="bn")(x)
    x = layers.Activation("relu", name="relu")(x)

    if include_top:
        x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
        x = layers.Dense(
            num_classes, activation=classifier_activation, name="predictions"
        )(x)
    elif pooling == "avg":
        x = layers.GlobalAveragePooling2D(name="avg_pool")(x)
    elif pooling == "max":
        x = layers.GlobalMaxPooling2D(name="max_pool")(x)

    model = keras.Model(inputs, x, name=name, **kwargs)

    if weights is not None:
        model.load_weights(weights)
    return model


def DenseNet121(
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_shape=(None, None, 3),
    pooling=None,
    name="DenseNet121",
    **kwargs,
):
    return DenseNet(
        [6, 12, 24, 16],
        include_rescaling=include_rescaling,
        include_top=include_top,
        num_classes=num_classes,
        weights=weights,
        input_shape=input_shape,
        pooling=pooling,
        name=name,
        **kwargs,
    )


def DenseNet169(
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_shape=(None, None, 3),
    pooling=None,
    name="DenseNet169",
    **kwargs,
):
    return DenseNet(
        [6, 12, 32, 32],
        include_rescaling=include_rescaling,
        include_top=include_top,
        num_classes=num_classes,
        weights=weights,
        input_shape=input_shape,
        pooling=pooling,
        name=name,
        **kwargs,
    )


def DenseNet201(
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_shape=(None, None, 3),
    pooling=None,
    name="DenseNet201",
    **kwargs,
):
    return DenseNet(
        blocks=[6, 12, 48, 32],
        include_rescaling=include_rescaling,
        include_top=include_top,
        num_classes=num_classes,
        weights=weights,
        input_shape=input_shape,
        pooling=pooling,
        name=name,
        **kwargs,
    )


setattr(DenseNet121, "__doc__", BASE_DOCSTRING.format(name="DenseNet121"))
setattr(DenseNet169, "__doc__", BASE_DOCSTRING.format(name="DenseNet169"))
setattr(DenseNet201, "__doc__", BASE_DOCSTRING.format(name="DenseNet201"))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import preprocessing


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class Solarization(BaseImageAugmentationLayer):
    """Applies (max_value - pixel + min_value) for each pixel in the image.

    When created without `threshold` parameter, the layer performs solarization to
    all values. When created with specified `threshold` the layer only augments
    pixels that are above the `threshold` value

    Reference:
    - [AutoAugment: Learning Augmentation Policies from Data](
        https://arxiv.org/abs/1805.09501
    )
    - [RandAugment](https://arxiv.org/pdf/1909.13719.pdf)

    Args:
        value_range: a tuple or a list of two elements. The first value represents
            the lower bound for values in passed images, the second represents the
            upper bound. Images passed to the layer should have values within
            `value_range`.
        addition_factor: (Optional)  A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is sampled
            from the provided range. If a float is passed, the range is interpreted as
            `(0, addition_factor)`. If specified, this value is added to each pixel
            before solarization and thresholding.  The addition value should be scaled
            according to the value range (0, 255). Defaults to 0.0.
        threshold_factor: (Optional)  A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is sampled
            from the provided range. If a float is passed, the range is interpreted as
            `(0, threshold_factor)`. If specified, only pixel values above this
            threshold will be solarized.
        seed: Integer. Used to create a random seed.

    Usage:
    ```python
    (images, labels), _ = tf.keras.datasets.cifar10.load_data()
    print(images[0, 0, 0])
    # [59 62 63]
    # Note that images are Tensor with values in the range [0, 255]
    solarization = Solarization()
    images = solarization(images)
    print(images[0, 0, 0])
    # [196, 193, 192]
    ```

    Call arguments:
        images: Tensor of type int or float, with pixels in
            range [0, 255] and shape [batch, height, width, channels]
            or [height, width, channels].
    """

    def __init__(
        self,
        value_range,
        addition_factor=0.0,
        threshold_factor=0.0,
        seed=None,
        **kwargs
    ):
        super().__init__(seed=seed, **kwargs)
        self.addition_factor = preprocessing.parse_factor(
            addition_factor, max_value=255, seed=seed, param_name="addition_factor"
        )
        self.threshold_factor = preprocessing.parse_factor(
            threshold_factor, max_value=255, seed=seed, param_name="threshold_factor"
        )
        self.value_range = value_range

    def get_random_transformation(self, image=None, label=None, bounding_boxes=None):
        return (self.addition_factor(), self.threshold_factor())

    def augment_image(self, image, transformation=None, **kwargs):
        (addition, threshold) = transformation
        image = preprocessing.transform_value_range(
            image, original_range=self.value_range, target_range=(0, 255)
        )
        result = image + addition
        result = tf.clip_by_value(result, 0, 255)
        result = tf.where(result < threshold, result, 255 - result)
        result = preprocessing.transform_value_range(
            result, original_range=(0, 255), target_range=self.value_range
        )
        return result

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    def get_config(self):
        config = {
            "threshold_factor": self.threshold_factor,
            "addition_factor": self.addition_factor,
            "value_range": self.value_range,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers import preprocessing


class RandomColorDegenerationTest(tf.test.TestCase):
    def test_random_color_degeneration_base_case(self):
        img_shape = (50, 50, 3)
        xs = tf.stack(
            [2 * tf.ones(img_shape), tf.ones(img_shape)],
            axis=0,
        )

        layer = preprocessing.RandomColorDegeneration(0.0)
        ys = layer(xs)

        self.assertEqual(xs.shape, ys.shape)

    def test_color_degeneration_full_factor(self):
        img_shape = (50, 50, 1)
        r = tf.ones(img_shape)
        g = 2 * tf.ones(img_shape)
        b = 3 * tf.ones(img_shape)
        xs = tf.concat([r, g, b], axis=-1)

        layer = preprocessing.RandomColorDegeneration(factor=(1, 1))
        ys = layer(xs)

        # Color degeneration uses standard luma conversion for RGB->Grayscale.
        # The formula for luma is result= 0.2989*r + 0.5870*g + 0.1140*b
        luma_result = 0.2989 + 2 * 0.5870 + 3 * 0.1140
        self.assertAllClose(ys, tf.ones_like(ys) * luma_result)

    def test_color_degeneration_70p_factor(self):
        img_shape = (50, 50, 1)
        r = tf.ones(img_shape)
        g = 2 * tf.ones(img_shape)
        b = 3 * tf.ones(img_shape)
        xs = tf.concat([r, g, b], axis=-1)

        layer = preprocessing.RandomColorDegeneration(factor=(0.7, 0.7))
        ys = layer(xs)

        # Color degeneration uses standard luma conversion for RGB->Grayscale.
        # The formula for luma is result= 0.2989*r + 0.5870*g + 0.1140*b
        luma_result = 0.2989 + 2 * 0.5870 + 3 * 0.1140

        # with factor=0.7, luma_result should be blended at a 70% rate with the original
        r_result = luma_result * 0.7 + 1 * 0.3
        g_result = luma_result * 0.7 + 2 * 0.3
        b_result = luma_result * 0.7 + 3 * 0.3

        r = ys[..., 0]
        g = ys[..., 1]
        b = ys[..., 2]

        self.assertAllClose(r, tf.ones_like(r) * r_result)
        self.assertAllClose(g, tf.ones_like(g) * g_result)
        self.assertAllClose(b, tf.ones_like(b) * b_result)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl.testing import parameterized

from keras_cv import layers


class RandAugmentTest(tf.test.TestCase, parameterized.TestCase):
    @parameterized.named_parameters(
        ("0", 0),
        ("20", 0.2),
        ("55", 0.55),
        ("10", 1.0),
    )
    def test_runs_with_magnitude(self, magnitude):
        rand_augment = layers.RandAugment(
            value_range=(0, 255), rate=0.5, magnitude=magnitude
        )
        xs = tf.ones((2, 512, 512, 3))
        ys = rand_augment(xs)
        self.assertEqual(ys.shape, (2, 512, 512, 3))

    @parameterized.named_parameters(
        ("0_255", 0, 255),
        ("neg_1_1", -1, 1),
        ("0_1", 0, 1),
    )
    def test_runs_with_value_range(self, low, high):
        rand_augment = layers.RandAugment(
            augmentations_per_image=3, magnitude=0.5, rate=1.0, value_range=(low, high)
        )
        xs = tf.random.uniform((2, 512, 512, 3), low, high, dtype=tf.float32)
        ys = rand_augment(xs)
        self.assertTrue(tf.math.reduce_all(tf.logical_and(ys >= low, ys <= high)))

    @parameterized.named_parameters(
        ("float32", tf.float32),
        ("int32", tf.int32),
        ("uint8", tf.uint8),
    )
    def test_runs_with_dtype_input(self, dtype):
        rand_augment = layers.RandAugment(value_range=(0, 255))
        xs = tf.ones((2, 512, 512, 3), dtype=dtype)
        ys = rand_augment(xs)
        self.assertEqual(ys.shape, (2, 512, 512, 3))

    @parameterized.named_parameters(
        ("0_255", 0, 255),
        ("neg1_1", -1, 1),
        ("0_1", 0, 1),
    )
    def test_standard_policy_respects_value_range(self, lower, upper):
        my_layers = layers.RandAugment.get_standard_policy(
            value_range=(lower, upper), magnitude=1.0, magnitude_stddev=0.2
        )
        rand_augment = layers.RandomAugmentationPipeline(
            layers=my_layers, augmentations_per_image=3
        )
        xs = tf.random.uniform((2, 512, 512, 3), lower, upper, dtype=tf.float32)
        ys = rand_augment(xs)
        self.assertLessEqual(tf.math.reduce_max(ys), upper)
        self.assertGreaterEqual(tf.math.reduce_min(ys), lower)

    def test_runs_unbatched(self):
        rand_augment = layers.RandAugment(
            augmentations_per_image=3, magnitude=0.5, rate=1.0, value_range=(0, 255)
        )
        xs = tf.random.uniform((512, 512, 3), 0, 255, dtype=tf.float32)
        ys = rand_augment(xs)
        self.assertEqual(xs.shape, ys.shape)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.layers.regularization.dropblock_2d import DropBlock2D


class DropBlock2DTest(tf.test.TestCase):
    FEATURE_SHAPE = (1, 14, 14, 256)  # Shape of ResNet block group 3
    rng = tf.random.Generator.from_non_deterministic_state()

    def test_layer_not_created_with_invalid_block_size(self):
        invalid_sizes = [0, -10, (5, -2), (0, 7), (1, 2, 3, 4)]
        for size in invalid_sizes:
            with self.assertRaises(ValueError):
                DropBlock2D(block_size=size, rate=0.1)

    def test_layer_not_created_with_invalid_rate(self):
        invalid_rates = [1.1, -0.1]
        for rate in invalid_rates:
            with self.assertRaises(ValueError):
                DropBlock2D(rate=rate, block_size=7)

    def test_input_unchanged_in_eval_mode(self):
        dummy_inputs = self.rng.uniform(shape=self.FEATURE_SHAPE)
        layer = DropBlock2D(rate=0.1, block_size=7)

        output = layer(dummy_inputs, training=False)

        self.assertAllClose(dummy_inputs, output)

    def test_input_unchanged_with_rate_equal_to_zero(self):
        dummy_inputs = self.rng.uniform(shape=self.FEATURE_SHAPE)
        layer = DropBlock2D(rate=0.0, block_size=7)

        output = layer(dummy_inputs, training=True)

        self.assertAllClose(dummy_inputs, output)

    def test_input_gets_partially_zeroed_out_in_train_mode(self):
        dummy_inputs = self.rng.uniform(shape=self.FEATURE_SHAPE)
        layer = DropBlock2D(rate=0.1, block_size=7)

        output = layer(dummy_inputs, training=True)
        num_input_zeros = self._count_zeros(dummy_inputs)
        num_output_zeros = self._count_zeros(output)

        self.assertGreater(num_output_zeros, num_input_zeros)

    def test_batched_input_gets_partially_zeroed_out_in_train_mode(self):
        batched_shape = (4, *self.FEATURE_SHAPE[1:])
        dummy_inputs = self.rng.uniform(shape=batched_shape)
        layer = DropBlock2D(rate=0.1, block_size=7)

        output = layer(dummy_inputs, training=True)
        num_input_zeros = self._count_zeros(dummy_inputs)
        num_output_zeros = self._count_zeros(output)

        self.assertGreater(num_output_zeros, num_input_zeros)

    def test_input_gets_partially_zeroed_out_with_non_square_block_size(self):
        dummy_inputs = self.rng.uniform(shape=self.FEATURE_SHAPE)
        layer = DropBlock2D(rate=0.1, block_size=(7, 10))

        output = layer(dummy_inputs, training=True)
        num_input_zeros = self._count_zeros(dummy_inputs)
        num_output_zeros = self._count_zeros(output)

        self.assertGreater(num_output_zeros, num_input_zeros)

    @staticmethod
    def _count_zeros(tensor: tf.Tensor) -> tf.Tensor:
        return tf.size(tensor) - tf.math.count_nonzero(tensor, dtype=tf.int32)

    def test_works_with_xla(self):
        dummy_inputs = self.rng.uniform(shape=self.FEATURE_SHAPE)
        layer = DropBlock2D(rate=0.1, block_size=7)

        @tf.function(jit_compile=True)
        def apply(x):
            return layer(x, training=True)

        apply(dummy_inputs)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers import preprocessing


class RandomCutoutTest(tf.test.TestCase):
    def _run_test(self, height_factor, width_factor):
        img_shape = (40, 40, 3)
        xs = tf.stack(
            [2 * tf.ones(img_shape), tf.ones(img_shape)],
            axis=0,
        )
        xs = tf.cast(xs, tf.float32)

        fill_value = 0.0
        layer = preprocessing.RandomCutout(
            height_factor=height_factor,
            width_factor=width_factor,
            fill_mode="constant",
            fill_value=fill_value,
            seed=1,
        )
        xs = layer(xs)

        # Some pixels should be replaced with fill value
        self.assertTrue(tf.math.reduce_any(xs[0] == fill_value))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == fill_value))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_return_shapes(self):
        xs = tf.ones((2, 512, 512, 3))

        layer = preprocessing.RandomCutout(height_factor=0.5, width_factor=0.5, seed=1)
        xs = layer(xs)

        self.assertEqual(xs.shape, [2, 512, 512, 3])

    def test_return_shapes_single_element(self):
        xs = tf.ones((512, 512, 3))

        layer = preprocessing.RandomCutout(height_factor=0.5, width_factor=0.5, seed=1)
        xs = layer(xs)

        self.assertEqual(xs.shape, [512, 512, 3])

    def test_random_cutout_single_float(self):
        self._run_test(0.5, 0.5)

    def test_random_cutout_tuple_float(self):
        self._run_test((0.4, 0.9), (0.1, 0.3))

    def test_random_cutout_fail_mix_bad_param_values(self):
        fn = lambda: self._run_test(0.5, (15.0, 30))
        self.assertRaises(ValueError, fn)

    def test_random_cutout_fail_reverse_lower_upper_float(self):
        fn = lambda: self._run_test(0.5, (0.9, 0.4))
        self.assertRaises(ValueError, fn)

    def test_random_cutout_call_results_one_channel(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((40, 40, 1)), tf.ones((40, 40, 1))],
                axis=0,
            ),
            tf.float32,
        )

        patch_value = 0.0
        layer = preprocessing.RandomCutout(
            height_factor=0.5,
            width_factor=0.5,
            fill_mode="constant",
            fill_value=patch_value,
            seed=1,
        )
        xs = layer(xs)

        # Some pixels should be replaced with fill value
        self.assertTrue(tf.math.reduce_any(xs[0] == patch_value))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == patch_value))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_random_cutout_call_tiny_image(self):
        img_shape = (4, 4, 3)
        xs = tf.stack(
            [2 * tf.ones(img_shape), tf.ones(img_shape)],
            axis=0,
        )
        xs = tf.cast(xs, tf.float32)

        fill_value = 0.0
        layer = preprocessing.RandomCutout(
            height_factor=(0.4, 0.9),
            width_factor=(0.1, 0.3),
            fill_mode="constant",
            fill_value=fill_value,
            seed=1,
        )
        xs = layer(xs)

        # Some pixels should be replaced with fill value
        self.assertTrue(tf.math.reduce_any(xs[0] == fill_value))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == fill_value))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_in_tf_function(self):
        xs = tf.cast(
            tf.stack([2 * tf.ones((100, 100, 1)), tf.ones((100, 100, 1))], axis=0),
            tf.float32,
        )

        patch_value = 0.0
        layer = preprocessing.RandomCutout(
            height_factor=0.5,
            width_factor=0.5,
            fill_mode="constant",
            fill_value=patch_value,
            seed=1,
        )

        @tf.function
        def augment(x):
            return layer(x)

        xs = augment(xs)

        # Some pixels should be replaced with fill value
        self.assertTrue(tf.math.reduce_any(xs[0] == patch_value))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == patch_value))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import tensorflow as tf

import keras_cv
from keras_cv.layers.preprocessing.grid_mask import GridMask


class GridMaskTest(tf.test.TestCase):
    def test_return_shapes(self):
        xs = tf.ones((2, 512, 512, 3))

        layer = GridMask(ratio_factor=0.1, rotation_factor=(-0.2, 0.3))
        xs = layer(xs, training=True)

        self.assertEqual(xs.shape, [2, 512, 512, 3])

    def test_gridmask_call_results_one_channel(self):
        xs = tf.cast(
            tf.stack(
                [3 * tf.ones((40, 40, 1)), 2 * tf.ones((40, 40, 1))],
                axis=0,
            ),
            dtype=tf.float32,
        )

        fill_value = 0.0
        layer = GridMask(
            ratio_factor=0.3,
            rotation_factor=(0.2, 0.3),
            fill_mode="constant",
            fill_value=fill_value,
        )
        xs = layer(xs, training=True)

        # Some pixels should be replaced with fill_value
        self.assertTrue(tf.math.reduce_any(xs[0] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[0] == 3.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[1] == 2.0))

    def test_non_square_image(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((1024, 512, 1)), tf.ones((1024, 512, 1))],
                axis=0,
            ),
            dtype=tf.float32,
        )

        fill_value = 100.0
        layer = GridMask(
            ratio_factor=0.6,
            rotation_factor=0.3,
            fill_mode="constant",
            fill_value=fill_value,
        )
        xs = layer(xs, training=True)

        # Some pixels should be replaced with fill_value
        self.assertTrue(tf.math.reduce_any(xs[0] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_in_tf_function(self):
        xs = tf.cast(
            tf.stack([2 * tf.ones((100, 100, 1)), tf.ones((100, 100, 1))], axis=0),
            dtype=tf.float32,
        )

        fill_value = 255.0
        layer = GridMask(
            ratio_factor=keras_cv.ConstantFactorSampler(0.5),
            rotation_factor=0.5,
            fill_mode="constant",
            fill_value=fill_value,
        )

        @tf.function
        def augment(x):
            return layer(x, training=True)

        xs = augment(xs)

        # Some pixels should be replaced with fill_value
        self.assertTrue(tf.math.reduce_any(xs[0] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_in_single_image(self):
        xs = tf.cast(
            tf.ones((512, 512, 1)),
            dtype=tf.float32,
        )

        layer = GridMask(ratio_factor=(0.5, 0.5), fill_mode="constant", fill_value=0.0)
        xs = layer(xs, training=True)
        self.assertTrue(tf.math.reduce_any(xs == 0.0))
        self.assertTrue(tf.math.reduce_any(xs == 1.0))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl.testing import parameterized

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.layers.preprocessing.maybe_apply import MaybeApply


class ZeroOut(BaseImageAugmentationLayer):
    """Zero out all entries, for testing purposes."""

    def __init__(self):
        super(ZeroOut, self).__init__()

    def augment_image(self, image, transformation=None, **kwargs):
        return 0 * image

    def augment_label(self, label, transformation=None, **kwargs):
        return 0 * label


class MaybeApplyTest(tf.test.TestCase, parameterized.TestCase):
    rng = tf.random.Generator.from_seed(seed=1234)

    @parameterized.parameters([-0.5, 1.7])
    def test_raises_error_on_invalid_rate_parameter(self, invalid_rate):
        with self.assertRaises(ValueError):
            MaybeApply(rate=invalid_rate, layer=ZeroOut())

    def test_works_with_batched_input(self):
        batch_size = 32
        dummy_inputs = self.rng.uniform(shape=(batch_size, 224, 224, 3))
        layer = MaybeApply(rate=0.5, layer=ZeroOut(), seed=1234)

        outputs = layer(dummy_inputs)
        num_zero_inputs = self._num_zero_batches(dummy_inputs)
        num_zero_outputs = self._num_zero_batches(outputs)

        self.assertEqual(num_zero_inputs, 0)
        self.assertLess(num_zero_outputs, batch_size)
        self.assertGreater(num_zero_outputs, 0)

    @staticmethod
    def _num_zero_batches(images):
        num_batches = tf.shape(images)[0]
        num_non_zero_batches = tf.math.count_nonzero(
            tf.math.count_nonzero(images, axis=[1, 2, 3]), dtype=tf.int32
        )
        return num_batches - num_non_zero_batches

    def test_inputs_unchanged_with_zero_rate(self):
        dummy_inputs = self.rng.uniform(shape=(32, 224, 224, 3))
        layer = MaybeApply(rate=0.0, layer=ZeroOut())

        outputs = layer(dummy_inputs)

        self.assertAllClose(outputs, dummy_inputs)

    def test_all_inputs_changed_with_rate_equal_to_one(self):
        dummy_inputs = self.rng.uniform(shape=(32, 224, 224, 3))
        layer = MaybeApply(rate=1.0, layer=ZeroOut())

        outputs = layer(dummy_inputs)

        self.assertAllEqual(outputs, tf.zeros_like(dummy_inputs))

    def test_works_with_single_image(self):
        dummy_inputs = self.rng.uniform(shape=(224, 224, 3))
        layer = MaybeApply(rate=1.0, layer=ZeroOut())

        outputs = layer(dummy_inputs)

        self.assertAllEqual(outputs, tf.zeros_like(dummy_inputs))

    def test_can_modify_label(self):
        dummy_inputs = self.rng.uniform(shape=(32, 224, 224, 3))
        dummy_labels = tf.ones(shape=(32, 2))
        layer = MaybeApply(rate=1.0, layer=ZeroOut())

        outputs = layer({"images": dummy_inputs, "labels": dummy_labels})

        self.assertAllEqual(outputs["labels"], tf.zeros_like(dummy_labels))

    def test_works_with_native_keras_layers(self):
        dummy_inputs = self.rng.uniform(shape=(32, 224, 224, 3))
        zero_out = tf.keras.layers.Lambda(lambda x: {"images": 0 * x["images"]})
        layer = MaybeApply(rate=1.0, layer=zero_out)

        outputs = layer(dummy_inputs)

        self.assertAllEqual(outputs, tf.zeros_like(dummy_inputs))

    def test_works_with_xla(self):
        dummy_inputs = self.rng.uniform(shape=(32, 224, 224, 3))
        # auto_vectorize=True will crash XLA
        layer = MaybeApply(rate=0.5, layer=ZeroOut(), auto_vectorize=False)

        @tf.function(jit_compile=True)
        def apply(x):
            return layer(x)

        apply(dummy_inputs)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf
from absl.testing import parameterized

from keras_cv import layers


class AddOneToInputs(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def call(self, inputs):
        result = inputs.copy()
        result["images"] = inputs["images"] + 1
        return result


class RandomAugmentationPipelineTest(tf.test.TestCase, parameterized.TestCase):
    @parameterized.named_parameters(("1", 1), ("3", 3), ("5", 5))
    def test_calls_layers_augmentations_per_image_times(self, augmentations_per_image):
        layer = AddOneToInputs()
        pipeline = layers.RandomAugmentationPipeline(
            layers=[layer], augmentations_per_image=augmentations_per_image, rate=1.0
        )
        xs = tf.random.uniform((2, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs + augmentations_per_image, os)

    def test_calls_layers_augmentations_in_graph(self):
        layer = AddOneToInputs()
        pipeline = layers.RandomAugmentationPipeline(
            layers=[layer], augmentations_per_image=3, rate=1.0
        )

        @tf.function()
        def call_pipeline(xs):
            return pipeline(xs)

        xs = tf.random.uniform((2, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = call_pipeline(xs)

        self.assertAllClose(xs + 3, os)

    @parameterized.named_parameters(("1", 1), ("3", 3), ("5", 5))
    def test_calls_layers_augmentations_per_image_times_single_image(
        self, augmentations_per_image
    ):
        layer = AddOneToInputs()
        pipeline = layers.RandomAugmentationPipeline(
            layers=[layer], augmentations_per_image=augmentations_per_image, rate=1.0
        )
        xs = tf.random.uniform((5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs + augmentations_per_image, os)

    @parameterized.named_parameters(("1", 1), ("3", 3), ("5", 5))
    def test_respects_rate(self, augmentations_per_image):
        layer = AddOneToInputs()
        pipeline = layers.RandomAugmentationPipeline(
            layers=[layer], augmentations_per_image=augmentations_per_image, rate=0.0
        )
        xs = tf.random.uniform((2, 5, 5, 3), 0, 100, dtype=tf.float32)
        os = pipeline(xs)

        self.assertAllClose(xs, os)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf
from absl.testing import parameterized

from keras_cv.layers import preprocessing


class RandomChannelShiftTest(tf.test.TestCase, parameterized.TestCase):
    def test_return_shapes(self):
        xs = tf.ones((2, 512, 512, 3))
        layer = preprocessing.RandomChannelShift(factor=1.0, value_range=(0, 255))

        xs = layer(xs, training=True)
        self.assertEqual(xs.shape, [2, 512, 512, 3])

    def test_non_square_image(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((1024, 512, 3)), tf.ones((1024, 512, 3))],
                axis=0,
            ),
            dtype=tf.float32,
        )
        layer = preprocessing.RandomChannelShift(
            factor=[0.1, 0.3], value_range=(0, 255)
        )

        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs[0] == 2.0))
        self.assertFalse(tf.math.reduce_any(xs[1] == 1.0))

    def test_in_tf_function(self):
        xs = tf.cast(
            tf.stack([2 * tf.ones((100, 100, 3)), tf.ones((100, 100, 3))], axis=0),
            dtype=tf.float32,
        )
        layer = preprocessing.RandomChannelShift(factor=0.3, value_range=(0, 255))

        @tf.function
        def augment(x):
            return layer(x, training=True)

        xs = augment(xs)
        self.assertFalse(tf.math.reduce_any(xs[0] == 2.0))
        self.assertFalse(tf.math.reduce_any(xs[1] == 1.0))

    def test_5_channels(self):
        xs = tf.cast(
            tf.ones((512, 512, 5)),
            dtype=tf.float32,
        )
        layer = preprocessing.RandomChannelShift(
            factor=0.4, channels=5, value_range=(0, 255)
        )
        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs == 1.0))

    def test_1_channel(self):
        xs = tf.cast(
            tf.ones((512, 512, 1)),
            dtype=tf.float32,
        )
        layer = preprocessing.RandomChannelShift(
            factor=0.4, channels=1, value_range=(0, 255)
        )
        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs == 1.0))

    def test_in_single_image(self):
        xs = tf.cast(
            tf.ones((512, 512, 3)),
            dtype=tf.float32,
        )
        layer = preprocessing.RandomChannelShift(factor=0.4, value_range=(0, 255))
        xs = layer(xs, training=True)
        self.assertFalse(tf.math.reduce_any(xs == 1.0))

    def test_config(self):
        layer = preprocessing.RandomChannelShift(
            factor=[0.1, 0.5], value_range=(0, 255), seed=101
        )
        config = layer.get_config()
        self.assertEqual(config["factor"].get_config()["lower"], 0.1)
        self.assertEqual(config["factor"].get_config()["upper"], 0.5)
        self.assertEqual(config["value_range"], (0, 255))
        self.assertEqual(config["channels"], 3)
        self.assertEqual(config["seed"], 101)

        reconstructed_layer = preprocessing.RandomChannelShift.from_config(config)
        self.assertEqual(reconstructed_layer.factor, layer.factor)
        self.assertEqual(reconstructed_layer.value_range, layer.value_range)
        self.assertEqual(reconstructed_layer.seed, layer.seed)
        self.assertEqual(reconstructed_layer.channels, layer.channels)

    def test_inference(self):
        layer = preprocessing.RandomChannelShift(factor=0.8, value_range=(0, 255))
        inputs = np.random.randint(0, 255, size=(224, 224, 3))
        output = layer(inputs, training=False)
        self.assertAllClose(inputs, output)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import fill_utils


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class CutMix(BaseImageAugmentationLayer):
    """CutMix implements the CutMix data augmentation technique.

    Args:
        alpha: Float between 0 and 1.  Inverse scale parameter for the gamma
            distribution.  This controls the shape of the distribution from which the
            smoothing values are sampled.  Defaults 1.0, which is a recommended value
            when training an imagenet1k classification model.
        seed: Integer. Used to create a random seed.
    References:
       [CutMix paper]( https://arxiv.org/abs/1905.04899).

    Sample usage:
    ```python
    (images, labels), _ = tf.keras.datasets.cifar10.load_data()
    labels = tf.one_hot(labels.squeeze(), 10)

    cutmix = keras_cv.layers.preprocessing.cut_mix.CutMix(10)
    output = cutmix({"images": images[:32], "labels": labels[:32]})
    # output == {'images': updated_images, 'labels': updated_labels}
    ```
    """

    def __init__(self, alpha=1.0, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.alpha = alpha
        self.seed = seed

    @staticmethod
    def _sample_from_beta(alpha, beta, shape):
        sample_alpha = tf.random.gamma(shape, 1.0, beta=alpha)
        sample_beta = tf.random.gamma(shape, 1.0, beta=beta)
        return sample_alpha / (sample_alpha + sample_beta)

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        images = inputs.get("images", None)
        labels = inputs.get("labels", None)
        if images is None or labels is None:
            raise ValueError(
                "CutMix expects inputs in a dictionary with format "
                '{"images": images, "labels": labels}.'
                f"Got: inputs = {inputs}"
            )
        images, labels = self._update_labels(*self._cutmix(images, labels))
        inputs["images"] = images
        inputs["labels"] = labels
        return inputs

    def _augment(self, inputs):
        raise ValueError(
            "CutMix received a single image to `call`.  The layer relies on "
            "combining multiple examples, and as such will not behave as "
            "expected.  Please call the layer with 2 or more samples."
        )

    def _cutmix(self, images, labels):
        """Apply cutmix."""
        input_shape = tf.shape(images)
        batch_size, image_height, image_width = (
            input_shape[0],
            input_shape[1],
            input_shape[2],
        )

        permutation_order = tf.random.shuffle(tf.range(0, batch_size), seed=self.seed)
        lambda_sample = CutMix._sample_from_beta(self.alpha, self.alpha, (batch_size,))

        ratio = tf.math.sqrt(1 - lambda_sample)

        cut_height = tf.cast(
            ratio * tf.cast(image_height, dtype=tf.float32), dtype=tf.int32
        )
        cut_width = tf.cast(
            ratio * tf.cast(image_height, dtype=tf.float32), dtype=tf.int32
        )

        random_center_height = tf.random.uniform(
            shape=[batch_size], minval=0, maxval=image_height, dtype=tf.int32
        )
        random_center_width = tf.random.uniform(
            shape=[batch_size], minval=0, maxval=image_width, dtype=tf.int32
        )

        bounding_box_area = cut_height * cut_width
        lambda_sample = 1.0 - bounding_box_area / (image_height * image_width)
        lambda_sample = tf.cast(lambda_sample, dtype=tf.float32)

        images = fill_utils.fill_rectangle(
            images,
            random_center_width,
            random_center_height,
            cut_width,
            cut_height,
            tf.gather(images, permutation_order),
        )

        return images, labels, lambda_sample, permutation_order

    def _update_labels(self, images, labels, lambda_sample, permutation_order):
        cutout_labels = tf.gather(labels, permutation_order)

        lambda_sample = tf.reshape(lambda_sample, [-1, 1])
        labels = lambda_sample * labels + (1.0 - lambda_sample) * cutout_labels
        return images, labels

    def _validate_inputs(self, inputs):
        labels = inputs.get("labels", None)
        if labels is None:
            raise ValueError(
                "CutMix expects 'labels' to be present in its inputs. "
                "CutMix relies on both images an labels. "
                "Please pass a dictionary with keys 'images' "
                "containing the image Tensor, and 'labels' containing "
                "the classification labels. "
                "For example, `cut_mix({'images': images, 'labels': labels})`."
            )
        if not labels.dtype.is_floating:
            raise ValueError(
                f"CutMix received labels with type {labels.dtype}. "
                "Labels must be of type float."
            )

    def get_config(self):
        config = {
            "alpha": self.alpha,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import fill_utils
from keras_cv.utils import preprocessing


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class RandomCutout(BaseImageAugmentationLayer):
    """Randomly cut out rectangles from images and fill them.

    Args:
        height_factor: A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`.  `height_factor` controls the size of the
            cutouts. `height_factor=0.0` means the rectangle will be of size 0% of the
            image height, `height_factor=0.1` means the rectangle will have a size of
            10% of the image height, and so forth.
            Values should be between `0.0` and `1.0`.  If a tuple is used, a
            `height_factor` is sampled between the two values for every image augmented.
            If a single float is used, a value between `0.0` and the passed float is
            sampled.  In order to ensure the value is always the same, please pass a
            tuple with two identical floats: `(0.5, 0.5)`.
        width_factor: A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`.  `width_factor` controls the size of the
            cutouts. `width_factor=0.0` means the rectangle will be of size 0% of the
            image height, `width_factor=0.1` means the rectangle will have a size of 10%
            of the image width, and so forth.
            Values should be between `0.0` and `1.0`.  If a tuple is used, a
            `width_factor` is sampled between the two values for every image augmented.
            If a single float is used, a value between `0.0` and the passed float is
            sampled.  In order to ensure the value is always the same, please pass a
            tuple with two identical floats: `(0.5, 0.5)`.
        fill_mode: Pixels inside the patches are filled according to the given
            mode (one of `{"constant", "gaussian_noise"}`).
            - *constant*: Pixels are filled with the same constant value.
            - *gaussian_noise*: Pixels are filled with random gaussian noise.
        fill_value: a float represents the value to be filled inside the patches
            when `fill_mode="constant"`.
        seed: Integer. Used to create a random seed.

    Sample usage:
    ```python
    (images, labels), _ = tf.keras.datasets.cifar10.load_data()
    random_cutout = keras_cv.layers.preprocessing.RandomCutout(0.5, 0.5)
    augmented_images = random_cutout(images)
    ```
    """

    def __init__(
        self,
        height_factor,
        width_factor,
        fill_mode="constant",
        fill_value=0.0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)

        self.height_factor = preprocessing.parse_factor(
            height_factor, param_name="height_factor", seed=seed
        )
        self.width_factor = preprocessing.parse_factor(
            width_factor, param_name="width_factor", seed=seed
        )
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.seed = seed

        if fill_mode not in ["gaussian_noise", "constant"]:
            raise ValueError(
                '`fill_mode` should be "gaussian_noise" '
                f'or "constant".  Got `fill_mode`={fill_mode}'
            )

    def _parse_bounds(self, factor):
        if isinstance(factor, (tuple, list)):
            return factor[0], factor[1]
        else:
            return type(factor)(0), factor

    def get_random_transformation(self, image=None, label=None, bounding_boxes=None):
        center_x, center_y = self._compute_rectangle_position(image)
        rectangle_height, rectangle_width = self._compute_rectangle_size(image)
        return center_x, center_y, rectangle_height, rectangle_width

    def augment_image(self, image, transformation=None, **kwargs):
        """Apply random cutout."""
        inputs = tf.expand_dims(image, 0)
        center_x, center_y, rectangle_height, rectangle_width = transformation

        rectangle_fill = self._compute_rectangle_fill(inputs)
        inputs = fill_utils.fill_rectangle(
            inputs,
            center_x,
            center_y,
            rectangle_width,
            rectangle_height,
            rectangle_fill,
        )
        return inputs[0]

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    def _compute_rectangle_position(self, inputs):
        input_shape = tf.shape(inputs)
        image_height, image_width = (
            input_shape[0],
            input_shape[1],
        )
        center_x = self._random_generator.random_uniform(
            [1], 0, image_width, dtype=tf.int32
        )
        center_y = self._random_generator.random_uniform(
            [1], 0, image_height, dtype=tf.int32
        )
        return center_x, center_y

    def _compute_rectangle_size(self, inputs):
        input_shape = tf.shape(inputs)
        image_height, image_width = (
            input_shape[0],
            input_shape[1],
        )
        height = self.height_factor()
        width = self.width_factor()

        height = height * tf.cast(image_height, tf.float32)
        width = width * tf.cast(image_width, tf.float32)

        height = tf.cast(tf.math.ceil(height), tf.int32)
        width = tf.cast(tf.math.ceil(width), tf.int32)

        height = tf.minimum(height, image_height)
        width = tf.minimum(width, image_width)

        return tf.expand_dims(height, axis=0), tf.expand_dims(width, axis=0)

    def _compute_rectangle_fill(self, inputs):
        input_shape = tf.shape(inputs)
        if self.fill_mode == "constant":
            fill_value = tf.fill(input_shape, self.fill_value)
        else:
            # gaussian noise
            fill_value = tf.random.normal(input_shape)

        return fill_value

    def get_config(self):
        config = {
            "height_factor": self.height_factor,
            "width_factor": self.width_factor,
            "fill_mode": self.fill_mode,
            "fill_value": self.fill_value,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import warnings

import tensorflow as tf

from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import preprocessing


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class RandomShear(BaseImageAugmentationLayer):
    """Randomly shears an image.

    Args:
        x_factor: A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is sampled
            from the provided range. If a float is passed, the range is interpreted as
            `(0, x_factor)`.  Values represent a percentage of the image to shear over.
             For example, 0.3 shears pixels up to 30% of the way across the image.
             All provided values should be positive.  If `None` is passed, no shear
             occurs on the X axis.
             Defaults to `None`.
        y_factor: A tuple of two floats, a single float or a
            `keras_cv.FactorSampler`. For each augmented image a value is sampled
            from the provided range. If a float is passed, the range is interpreted as
            `(0, y_factor)`. Values represent a percentage of the image to shear over.
            For example, 0.3 shears pixels up to 30% of the way across the image.
            All provided values should be positive.  If `None` is passed, no shear
            occurs on the Y axis.
            Defaults to `None`.
        interpolation: interpolation method used in the `ImageProjectiveTransformV3` op.
             Supported values are `"nearest"` and `"bilinear"`.
             Defaults to `"bilinear"`.
        fill_mode: fill_mode in the `ImageProjectiveTransformV3` op.
             Supported values are `"reflect"`, `"wrap"`, `"constant"`, and `"nearest"`.
             Defaults to `"reflect"`.
        fill_value: fill_value in the `ImageProjectiveTransformV3` op.
             A `Tensor` of type `float32`. The value to be filled when fill_mode is
             constant".  Defaults to `0.0`.
        seed: Integer. Used to create a random seed.
    """

    def __init__(
        self,
        x_factor=None,
        y_factor=None,
        interpolation="bilinear",
        fill_mode="reflect",
        fill_value=0.0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if x_factor is not None:
            self.x_factor = preprocessing.parse_factor(
                x_factor, max_value=None, param_name="x_factor", seed=seed
            )
        else:
            self.x_factor = x_factor
        if y_factor is not None:
            self.y_factor = preprocessing.parse_factor(
                y_factor, max_value=None, param_name="y_factor", seed=seed
            )
        else:
            self.y_factor = y_factor
        if x_factor is None and y_factor is None:
            warnings.warn(
                "RandomShear received both `x_factor=None` and `y_factor=None`.  As a "
                "result, the layer will perform no augmentation."
            )
        self.interpolation = interpolation
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.seed = seed

    def get_random_transformation(self, image=None, label=None, bounding_boxes=None):
        x = self._get_shear_amount(self.x_factor)
        y = self._get_shear_amount(self.y_factor)
        return (x, y)

    def _get_shear_amount(self, constraint):
        if constraint is None:
            return None

        invert = preprocessing.random_inversion(self._random_generator)
        return invert * constraint()

    def augment_image(self, image, transformation=None, **kwargs):
        image = tf.expand_dims(image, axis=0)

        x, y = transformation

        if x is not None:
            transform_x = RandomShear._format_transform(
                [1.0, x, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0]
            )
            image = preprocessing.transform(
                images=image,
                transforms=transform_x,
                interpolation=self.interpolation,
                fill_mode=self.fill_mode,
                fill_value=self.fill_value,
            )

        if y is not None:
            transform_y = RandomShear._format_transform(
                [1.0, 0.0, 0.0, y, 1.0, 0.0, 0.0, 0.0]
            )
            image = preprocessing.transform(
                images=image,
                transforms=transform_y,
                interpolation=self.interpolation,
                fill_mode=self.fill_mode,
                fill_value=self.fill_value,
            )

        return tf.squeeze(image, axis=0)

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    @staticmethod
    def _format_transform(transform):
        transform = tf.convert_to_tensor(transform, dtype=tf.float32)
        return transform[tf.newaxis]

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "x_factor": self.x_factor,
                "y_factor": self.y_factor,
                "interpolation": self.interpolation,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "seed": self.seed,
            }
        )
        return config



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv.layers.regularization.stochastic_depth import StochasticDepth


class StochasticDepthTest(tf.test.TestCase):
    FEATURE_SHAPE = (1, 14, 14, 256)

    def test_inputs_have_two_elements(self):
        inputs = tf.random.uniform(self.FEATURE_SHAPE, 0, 1)
        inputs = [inputs, inputs, inputs]

        with self.assertRaisesRegex(
            ValueError, "Input must be a list of length 2. " "Got input with length=3."
        ):
            StochasticDepth()(inputs)

    def test_eval_mode(self):
        inputs = tf.random.uniform(self.FEATURE_SHAPE, 0, 1)
        inputs = [inputs, inputs]

        rate = 0.5

        outputs = StochasticDepth(rate=rate)(inputs, training=False)

        self.assertAllClose(inputs[0] * (1 + rate), outputs)

    def test_training_mode(self):
        inputs = tf.random.uniform(self.FEATURE_SHAPE, 0, 1)
        inputs = [inputs, inputs]

        rate = 0.5

        outputs = StochasticDepth(rate=rate)(inputs, training=True)

        outputs_sum = tf.math.reduce_sum(outputs)
        inputs_sum = tf.math.reduce_sum(inputs[0])

        self.assertIn(outputs_sum, [inputs_sum, inputs_sum * 2])



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.utils import preprocessing


class MockRandomGenerator:
    def __init__(self, value):
        self.value = value

    def random_uniform(self, shape, minval, maxval, dtype=None):
        del minval, maxval
        return tf.constant(self.value, dtype=dtype)


class PreprocessingTestCase(tf.test.TestCase):
    def setUp(self):
        super().setUp()

    def test_transform_to_standard_range_neg_one_range(self):
        x = tf.constant([-1, 0, 1])
        x = preprocessing.transform_value_range(
            x, original_range=[-1, 1], target_range=[0, 255]
        )
        self.assertAllClose(x, [0.0, 127.5, 255.0])

    def test_transform_to_same_range(self):
        x = tf.constant([-1, 0, 1])
        x = preprocessing.transform_value_range(
            x, original_range=[0, 255], target_range=[0, 255]
        )
        self.assertAllClose(x, [-1, 0, 1])

    def test_transform_to_standard_range(self):
        x = tf.constant([8 / 255, 9 / 255, 255 / 255])
        x = preprocessing.transform_value_range(
            x, original_range=[0, 1], target_range=[0, 255]
        )
        self.assertAllClose(x, [8.0, 9.0, 255.0])

    def test_transform_to_value_range(self):
        x = tf.constant([128.0, 255.0, 0.0])
        x = preprocessing.transform_value_range(
            x, original_range=[0, 255], target_range=[0, 1]
        )
        self.assertAllClose(x, [128 / 255, 1, 0])

    def test_random_inversion(self):
        generator = MockRandomGenerator(0.75)
        self.assertEqual(preprocessing.random_inversion(generator), -1.0)
        generator = MockRandomGenerator(0.25)
        self.assertEqual(preprocessing.random_inversion(generator), 1.0)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from keras_cv import core
from keras_cv.layers import preprocessing as cv_preprocessing
from keras_cv.layers.preprocessing.random_augmentation_pipeline import (
    RandomAugmentationPipeline,
)
from keras_cv.utils import preprocessing as preprocessing_utils


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class RandAugment(RandomAugmentationPipeline):
    """RandAugment performs the Rand Augment operation on input images.

    This layer can be thought of as an all in one image augmentation layer.  The policy
    implemented by this layer has been benchmarked extensively and is effective on a
    wide variety of datasets.

    The policy operates as follows:

    For each augmentation in the range `[0, augmentations_per_image]`,
    the policy selects a random operation from a list of operations.
    It then samples a random number and if that number is less than
    `rate` applies it to the given image.

    References:
        - [RandAugment](https://arxiv.org/abs/1909.13719)

    Args:
        value_range: the range of values the incoming images will have.
            Represented as a two number tuple written [low, high].
            This is typically either `[0, 1]` or `[0, 255]` depending
            on how your preprocessing pipeline is setup.
        augmentations_per_image: the number of layers to use in the rand augment policy.
            Defaults to `3`.
        magnitude: magnitude is the mean of the normal distribution used to sample the
            magnitude used for each data augmentation.  magnitude should
            be a float in the range `[0, 1]`.  A magnitude of `0` indicates that the
            augmentations are as weak as possible (not recommended), while a value of
            `1.0` implies use of the strongest possible augmentation.  All magnitudes
            are clipped to the range `[0, 1]` after sampling.  Defaults to `0.5`.
        magnitude_stddev: the standard deviation to use when drawing values
            for the perturbations.  Keep in mind magnitude will still be clipped to the
            range `[0, 1]` after samples are drawn from the normal distribution.
            Defaults to `0.15`.
        rate:  the rate at which to apply each augmentation.  This parameter is applied
            on a per-distortion layer, per image.  Should be in the range `[0, 1]`.
            To reproduce the original RandAugment paper results, set this to `10/11`.
            The original `RandAugment` paper includes an Identity transform.  By setting
            the rate to 10/11 in our implementation, the behavior is identical to
            sampling an Identity augmentation 10/11th of the time.
            Defaults to `1.0`.

    Usage:
    ```python
    (x_test, y_test), _ = tf.keras.datasets.cifar10.load_data()
    rand_augment = keras_cv.layers.RandAugment(
        value_range=(0, 255), augmentations_per_image=3, magnitude=0.5
    )
    x_test = rand_augment(x_test)
    ```
    """

    def __init__(
        self,
        value_range,
        augmentations_per_image=3,
        magnitude=0.5,
        magnitude_stddev=0.15,
        rate=10 / 11,
        seed=None,
        **kwargs,
    ):
        # As an optimization RandAugment makes all internal layers use (0, 255) while
        # and we handle range transformation at the _augment level.
        if magnitude < 0.0 or magnitude > 1:
            raise ValueError(
                f"`magnitude` must be in the range [0, 1], got `magnitude={magnitude}`"
            )
        if magnitude_stddev < 0.0 or magnitude_stddev > 1:
            raise ValueError(
                "`magnitude_stddev` must be in the range [0, 1], got "
                f"`magnitude_stddev={magnitude}`"
            )

        super().__init__(
            layers=RandAugment.get_standard_policy(
                (0, 255), magnitude, magnitude_stddev, seed=seed
            ),
            augmentations_per_image=augmentations_per_image,
            rate=rate,
            **kwargs,
            seed=seed,
        )
        self.magnitude = float(magnitude)
        self.value_range = value_range
        self.seed = seed
        self.magnitude_stddev = float(magnitude_stddev)

    def _augment(self, sample):
        sample["images"] = preprocessing_utils.transform_value_range(
            sample["images"], self.value_range, (0, 255)
        )
        result = super()._augment(sample)
        result["images"] = preprocessing_utils.transform_value_range(
            result["images"], (0, 255), self.value_range
        )
        result["images"]
        return result

    @staticmethod
    def get_standard_policy(value_range, magnitude, magnitude_stddev, seed=None):
        policy = create_rand_augment_policy(magnitude, magnitude_stddev)

        auto_contrast = cv_preprocessing.AutoContrast(
            **policy["auto_contrast"], value_range=value_range, seed=seed
        )
        equalize = cv_preprocessing.Equalization(
            **policy["equalize"], value_range=value_range, seed=seed
        )

        solarize = cv_preprocessing.Solarization(
            **policy["solarize"], value_range=value_range, seed=seed
        )

        color = cv_preprocessing.RandomColorDegeneration(**policy["color"], seed=seed)
        contrast = cv_preprocessing.RandomContrast(**policy["contrast"], seed=seed)
        brightness = cv_preprocessing.RandomBrightness(
            **policy["brightness"], value_range=value_range, seed=seed
        )
        shear_x = cv_preprocessing.RandomShear(**policy["shear_x"], seed=seed)
        shear_y = cv_preprocessing.RandomShear(**policy["shear_y"], seed=seed)
        translate_x = cv_preprocessing.RandomTranslation(
            **policy["translate_x"], seed=seed
        )
        translate_y = cv_preprocessing.RandomTranslation(
            **policy["translate_y"], seed=seed
        )
        return [
            auto_contrast,
            equalize,
            solarize,
            color,
            contrast,
            brightness,
            shear_x,
            shear_y,
            translate_x,
            translate_y,
        ]

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "augmentations_per_image": self.augmentations_per_image,
                "magnitude": self.magnitude,
                "magnitude_stddev": self.magnitude_stddev,
                "rate": self.rate,
                "seed": self.seed,
            }
        )
        # layers is recreated in the constructor
        del config["layers"]
        return config


def auto_contrast_policy(magnitude, magnitude_stddev):
    return {}


def equalize_policy(magnitude, magnitude_stddev):
    return {}


def solarize_policy(magnitude, magnitude_stddev):
    # We cap additions at 110, because if we add more than 110 we will be nearly
    # nullifying the information contained in the image, making the model train on noise
    maximum_addition_value = 110
    addition_factor = core.NormalFactorSampler(
        mean=magnitude * maximum_addition_value,
        stddev=magnitude_stddev * maximum_addition_value,
        min_value=0,
        max_value=maximum_addition_value,
    )
    threshold_factor = core.NormalFactorSampler(
        mean=(255 - (magnitude * 255)),
        stddev=(magnitude_stddev * 255),
        min_value=0,
        max_value=255,
    )
    return {"addition_factor": addition_factor, "threshold_factor": threshold_factor}


def color_policy(magnitude, magnitude_stddev):
    factor = core.NormalFactorSampler(
        mean=magnitude,
        stddev=magnitude_stddev,
        min_value=0,
        max_value=1,
    )
    return {"factor": factor}


def contrast_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomContrast with `factor`?
    # RandomContrast layer errors when factor=0
    factor = max(magnitude, 0.001)
    return {"factor": factor}


def brightness_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomBrightness with `factor`?
    return {"factor": magnitude}


def shear_x_policy(magnitude, magnitude_stddev):
    factor = core.NormalFactorSampler(
        mean=magnitude,
        stddev=magnitude_stddev,
        min_value=0,
        max_value=1,
    )
    return {"x_factor": factor, "y_factor": 0}


def shear_y_policy(magnitude, magnitude_stddev):
    factor = core.NormalFactorSampler(
        mean=magnitude,
        stddev=magnitude_stddev,
        min_value=0,
        max_value=1,
    )
    return {"x_factor": 0, "y_factor": factor}


def translate_x_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomTranslation with `factor`?
    return {"width_factor": magnitude, "height_factor": 0}


def translate_y_policy(magnitude, magnitude_stddev):
    # TODO(lukewood): should we integrate RandomTranslation with `factor`?
    return {"width_factor": 0, "height_factor": magnitude}


POLICY_PAIRS = {
    "auto_contrast": auto_contrast_policy,
    "equalize": equalize_policy,
    "solarize": solarize_policy,
    "color": color_policy,
    "contrast": contrast_policy,
    "brightness": brightness_policy,
    "shear_x": shear_x_policy,
    "shear_y": shear_y_policy,
    "translate_x": translate_x_policy,
    "translate_y": translate_y_policy,
}


def create_rand_augment_policy(magnitude, magnitude_stddev):
    result = {}
    for name, policy_fn in POLICY_PAIRS.items():
        result[name] = policy_fn(magnitude, magnitude_stddev)
    return result



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class DropPath(tf.keras.__internal__.layers.BaseRandomLayer):
    """
    Implements the DropPath layer. DropPath randomly drops samples during training
     with a probability of `rate`. Note that this layer drops individual samples
    within a batch and not the entire batch. DropPath randomly drops some of the
    individual samples from a batch, whereas StachasticDepth randomly drops the
    entire batch.

    References:
        - [FractalNet](https://arxiv.org/abs/1605.07648v4).
        - [rwightman/pytorch-image-models](https://github.com/rwightman/pytorch-image-models/blob/7c67d6aca992f039eece0af5f7c29a43d48c00e4/timm/models/layers/drop.py#L135)

    Args:
        rate: float, the probability of the residual branch being dropped.
        seed: (Optional) Integer. Used to create a random seed.

    Usage:
    `DropPath` can be used in any network as follows:
    ```python

    # (...)
    input = tf.ones((1, 3, 3, 1), dtype=tf.float32)
    residual = tf.keras.layers.Conv2D(1, 1)(input)
    output = keras_cv.layers.DropPath()(input)
    # (...)
    ```
    """

    def __init__(self, rate=0.5, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.rate = rate
        self.seed = seed

    def call(self, x, training=None):
        if self.rate == 0.0 or not training:
            return x
        else:
            keep_prob = 1 - self.rate
            drop_map_shape = (x.shape[0],) + (1,) * (len(x.shape) - 1)
            drop_map = tf.keras.backend.random_bernoulli(
                drop_map_shape, p=keep_prob, seed=self.seed
            )
            x = x / keep_prob
            x = x * drop_map
            return x

    def get_config(self):
        config = {"rate": self.rate, "seed": self.seed}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.layers import preprocessing


class AutoContrastTest(tf.test.TestCase):
    def test_constant_channels_dont_get_nanned(self):
        img = tf.constant([1, 1], dtype=tf.float32)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=0)

        layer = preprocessing.AutoContrast(value_range=(0, 255))
        ys = layer(img)

        self.assertTrue(tf.math.reduce_any(ys[0] == 1.0))
        self.assertTrue(tf.math.reduce_any(ys[0] == 1.0))

    def test_auto_contrast_expands_value_range(self):
        img = tf.constant([0, 128], dtype=tf.float32)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=0)

        layer = preprocessing.AutoContrast(value_range=(0, 255))
        ys = layer(img)

        self.assertTrue(tf.math.reduce_any(ys[0] == 0.0))
        self.assertTrue(tf.math.reduce_any(ys[0] == 255.0))

    def test_auto_contrast_different_values_per_channel(self):
        img = tf.constant(
            [[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]], dtype=tf.float32
        )
        img = tf.expand_dims(img, axis=0)

        layer = preprocessing.AutoContrast(value_range=(0, 255))
        ys = layer(img)

        self.assertTrue(tf.math.reduce_any(ys[0, ..., 0] == 0.0))
        self.assertTrue(tf.math.reduce_any(ys[0, ..., 1] == 0.0))

        self.assertTrue(tf.math.reduce_any(ys[0, ..., 0] == 255.0))
        self.assertTrue(tf.math.reduce_any(ys[0, ..., 1] == 255.0))

        self.assertAllClose(
            ys,
            [
                [
                    [[0.0, 0.0, 0.0], [85.0, 85.0, 85.0]],
                    [[170.0, 170.0, 170.0], [255.0, 255.0, 255.0]],
                ]
            ],
        )

    def test_auto_contrast_expands_value_range_uint8(self):
        img = tf.constant([0, 128], dtype=tf.uint8)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=0)

        layer = preprocessing.AutoContrast(value_range=(0, 255))
        ys = layer(img)

        self.assertTrue(tf.math.reduce_any(ys[0] == 0.0))
        self.assertTrue(tf.math.reduce_any(ys[0] == 255.0))

    def test_auto_contrast_properly_converts_value_range(self):
        img = tf.constant([0, 0.5], dtype=tf.float32)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=-1)
        img = tf.expand_dims(img, axis=0)

        layer = preprocessing.AutoContrast(value_range=(0, 1))
        ys = layer(img)

        self.assertTrue(tf.math.reduce_any(ys[0] == 0.0))
        self.assertTrue(tf.math.reduce_any(ys[0] == 1.0))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import tensorflow as tf

from keras_cv import bounding_box
from keras_cv.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.utils import preprocessing

# In order to support both unbatched and batched inputs, the horizontal
# and verticle axis is reverse indexed
H_AXIS = -3
W_AXIS = -2


@tf.keras.utils.register_keras_serializable(package="keras_cv")
class RandomRotation(BaseImageAugmentationLayer):
    """A preprocessing layer which randomly rotates images during training.

    This layer will apply random rotations to each image, filling empty space
    according to `fill_mode`.

    By default, random rotations are only applied during training.
    At inference time, the layer does nothing. If you need to apply random
    rotations at inference time, set `training` to True when calling the layer.

    Input pixel values can be of any range (e.g. `[0., 1.)` or `[0, 255]`) and
    of interger or floating point dtype. By default, the layer will output
    floats.

    Input shape:
      3D (unbatched) or 4D (batched) tensor with shape:
      `(..., height, width, channels)`, in `"channels_last"` format

    Output shape:
      3D (unbatched) or 4D (batched) tensor with shape:
      `(..., height, width, channels)`, in `"channels_last"` format

    Arguments:
      factor: a float represented as fraction of 2 Pi, or a tuple of size 2
        representing lower and upper bound for rotating clockwise and
        counter-clockwise. A positive values means rotating counter clock-wise,
        while a negative value means clock-wise. When represented as a single
        float, this value is used for both the upper and lower bound. For
        instance, `factor=(-0.2, 0.3)` results in an output rotation by a random
        amount in the range `[-20% * 2pi, 30% * 2pi]`. `factor=0.2` results in
        an output rotating by a random amount in the range
        `[-20% * 2pi, 20% * 2pi]`.
      fill_mode: Points outside the boundaries of the input are filled according
        to the given mode (one of `{"constant", "reflect", "wrap", "nearest"}`).
        - *reflect*: `(d c b a | a b c d | d c b a)` The input is extended by
          reflecting about the edge of the last pixel.
        - *constant*: `(k k k k | a b c d | k k k k)` The input is extended by
          filling all values beyond the edge with the same constant value k = 0.
        - *wrap*: `(a b c d | a b c d | a b c d)` The input is extended by
          wrapping around to the opposite edge.
        - *nearest*: `(a a a a | a b c d | d d d d)` The input is extended by
          the nearest pixel.
      interpolation: Interpolation mode. Supported values: `"nearest"`,
        `"bilinear"`.
      seed: Integer. Used to create a random seed.
      fill_value: a float represents the value to be filled outside the
        boundaries when `fill_mode="constant"`.
      bounding_box_format: The format of bounding boxes of input dataset. Refer
        https://github.com/keras-team/keras-cv/blob/master/keras_cv/bounding_box/converters.py
        for more details on supported bounding box formats.
    """

    def __init__(
        self,
        factor,
        fill_mode="reflect",
        interpolation="bilinear",
        seed=None,
        fill_value=0.0,
        bounding_box_format=None,
        **kwargs,
    ):
        super().__init__(seed=seed, force_generator=True, **kwargs)
        self.factor = factor
        if isinstance(factor, (tuple, list)):
            self.lower = factor[0]
            self.upper = factor[1]
        else:
            self.lower = -factor
            self.upper = factor
        if self.upper < self.lower:
            raise ValueError(
                "Factor cannot have negative values, " "got {}".format(factor)
            )
        preprocessing.check_fill_mode_and_interpolation(fill_mode, interpolation)
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.interpolation = interpolation
        self.seed = seed
        self.bounding_box_format = bounding_box_format

    def get_random_transformation(self, image=None, label=None, bounding_boxes=None):
        min_angle = self.lower * 2.0 * np.pi
        max_angle = self.upper * 2.0 * np.pi
        angle = self._random_generator.random_uniform(
            shape=[1], minval=min_angle, maxval=max_angle
        )
        return {"angle": angle}

    def augment_image(self, image, transformation, **kwargs):
        image = preprocessing.ensure_tensor(image, self.compute_dtype)
        original_shape = image.shape
        image = tf.expand_dims(image, 0)
        image_shape = tf.shape(image)
        img_hd = tf.cast(image_shape[H_AXIS], tf.float32)
        img_wd = tf.cast(image_shape[W_AXIS], tf.float32)
        angle = transformation["angle"]
        output = preprocessing.transform(
            image,
            preprocessing.get_rotation_matrix(angle, img_hd, img_wd),
            fill_mode=self.fill_mode,
            fill_value=self.fill_value,
            interpolation=self.interpolation,
        )
        output = tf.squeeze(output, 0)
        output.set_shape(original_shape)
        return output

    def augment_bounding_boxes(self, bounding_boxes, transformation, **kwargs):
        image = None
        image = kwargs.get("image")
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomRotation()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomRotation(bounding_box_format='xyxy')`"
            )
        else:
            bounding_boxes = bounding_box.convert_format(
                bounding_boxes, source=self.bounding_box_format, target="xyxy"
            )
        image = tf.expand_dims(image, 0)
        image_shape = tf.shape(image)
        h = image_shape[H_AXIS]
        w = image_shape[W_AXIS]
        # origin coordinates, all the points on the image are rotated around
        # this point
        origin_x, origin_y = int(h / 2), int(w / 2)
        angle = transformation["angle"]
        angle = -angle
        # calculate coordinates of all four corners of the bounding box
        point = tf.stack(
            [
                tf.stack([bounding_boxes[:, 0], bounding_boxes[:, 1]], axis=1),
                tf.stack([bounding_boxes[:, 2], bounding_boxes[:, 1]], axis=1),
                tf.stack([bounding_boxes[:, 2], bounding_boxes[:, 3]], axis=1),
                tf.stack([bounding_boxes[:, 0], bounding_boxes[:, 3]], axis=1),
            ],
            axis=1,
        )
        # point_x : x coordinates of all corners of the bounding box
        point_x = tf.gather(point, [0], axis=2)
        # point_y : y cordinates of all corners of the bounding box
        point_y = tf.gather(point, [1], axis=2)
        # rotated bounding box coordinates
        # new_x : new position of x coordinates of corners of bounding box
        new_x = (
            origin_x
            + tf.multiply(
                tf.cos(angle), tf.cast((point_x - origin_x), dtype=tf.float32)
            )
            - tf.multiply(
                tf.sin(angle), tf.cast((point_y - origin_y), dtype=tf.float32)
            )
        )
        # new_y : new position of y coordinates of corners of bounding box
        new_y = (
            origin_y
            + tf.multiply(
                tf.sin(angle), tf.cast((point_x - origin_x), dtype=tf.float32)
            )
            + tf.multiply(
                tf.cos(angle), tf.cast((point_y - origin_y), dtype=tf.float32)
            )
        )
        # rotated bounding box coordinates
        out = tf.concat([new_x, new_y], axis=2)
        # find readjusted coordinates of bounding box to represent it in corners
        # format
        min_cordinates = tf.math.reduce_min(out, axis=1)
        max_cordinates = tf.math.reduce_max(out, axis=1)
        bounding_boxes_out = tf.concat([min_cordinates, max_cordinates], axis=1)
        # cordinates cannot be float values, it is casted to int32
        bounding_boxes_out = bounding_box.convert_format(
            bounding_boxes_out,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
        )
        return bounding_boxes_out

    def augment_label(self, label, transformation, **kwargs):
        return label

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = {
            "factor": self.factor,
            "fill_mode": self.fill_mode,
            "fill_value": self.fill_value,
            "interpolation": self.interpolation,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))
